#include "iseghalremote.h"

#include "iseghalremotecommon.h"

#include <QSslSocket>

int IsegHalRemote::m_timeout = IsegHalRemote::DefaultNetworkTimeout;

IsegHalRemote::IsegHalRemote(QObject *parent) :
	QObject(parent)
{
	m_socket = new QSslSocket(this);
	m_socket->setProtocol(QSecureSslProtocols);
	m_socket->setPeerVerifyMode(QSslSocket::VerifyNone);
}

bool IsegHalRemote::connectToHostEncrypted(const QString &host, quint16 port)
{
	m_host = host.toLatin1();

	m_socket->connectToHostEncrypted(host, port);
	return m_socket->waitForEncrypted(timeout());
}

QByteArray IsegHalRemote::writeReadSynchronous(const QByteArray &data) const
{
	m_socket->write(data);
	if (!m_socket->waitForBytesWritten(timeout()))
		return QByteArray();

	m_socket->waitForReadyRead(timeout());
	return m_socket->readAll();
}

static QByteArray readResult(char type, const QByteArray &read, const QByteArray &session)
{
	const int sessionSize = session.size();
	if (read.mid(0, 1) != QByteArray(1, type))
		return QByteArray();
	if (read.mid(1, sessionSize) != session)
		return QByteArray();
	if (read.mid(1 + sessionSize, 1) != ";")
		return QByteArray();

	return read.mid(1 + sessionSize + 1);
}

QByteArray IsegHalRemote::writeRead(char command, const QByteArray &value) const
{
	QByteArray write = command + m_session;
	if (!value.isEmpty())
		write += ';' + value;
	const QByteArray read = writeReadSynchronous(write);
	const QByteArray data = readResult(command, read, m_session);

	return data;
}

int IsegHalRemote::timeout()
{
	return m_timeout;
}

void IsegHalRemote::setTimeout(int milliSeconds)
{
	m_timeout = milliSeconds;
}

bool IsegHalRemote::supportsSsl()
{
	return QSslSocket::supportsSsl();
}
IsegResult IsegHalRemote::isegConnect(const QString &session, const QString &interface)
{
	m_session = session.toLatin1();

	if (!supportsSsl())
		return ISEG_NO_SSL_SUPPORT;

	QRegExp reUrl;
	reUrl.setPattern(QLatin1String("^hal://(.*)/(.*)$"));
	reUrl.setMinimal(true);
	if (!reUrl.exactMatch(interface))
		return ISEG_ERROR;

	QRegExp rePort;
	rePort.setPattern(QLatin1String("(.*):(\\d{1,5})$"));
	rePort.setMinimal(true);

	quint16 port = ISEG_REMOTE_PORT;
	QString url = reUrl.cap(1);
	if (url.contains(rePort)) {
		bool ok;
		url = rePort.cap(1);
		port = rePort.cap(2).toUShort(&ok);

		if (!ok)
			return ISEG_ERROR;
	}

	if (!connectToHostEncrypted(url, port))
		return ISEG_ERROR;

	const QByteArray data = writeRead(ISEG_REMOTE_CONNECT, interface.toLatin1());
	if (data.isEmpty())
		return ISEG_ERROR;

	return static_cast<IsegResult>(data.toInt());
}

IsegResult IsegHalRemote::isegDisconnect()
{
	const QByteArray data = writeRead(ISEG_REMOTE_DISCONNECT);

	m_socket->disconnectFromHost();
	if (!m_socket->waitForDisconnected(timeout()))
		return ISEG_ERROR;

	if (data.isEmpty())
		return ISEG_ERROR;

	return static_cast<IsegResult>(data.toInt());
}

IsegItem IsegHalRemote::isegGetItem(const QByteArray &item)
{
	const QByteArray data = writeRead(ISEG_REMOTE_GETITEM, item);
	if (data.isEmpty())
		return IsegHalRemote::isegErrorItem();

	IsegItem result;
	memcpy(&result, data.constData(), sizeof(IsegItem));
	return result;
}

IsegResult IsegHalRemote::isegSetItem(const QByteArray &item, const QByteArray &value)
{
	const QByteArray data = writeRead(ISEG_REMOTE_SETITEM, item + "," + value);
	if (data.isEmpty())
		return ISEG_ERROR;

	return static_cast<IsegResult>(data.toInt());
}

IsegItemProperty IsegHalRemote::isegGetItemProperty(const QByteArray &item)
{
	const QByteArray data = writeRead(ISEG_REMOTE_GETPROPERTY, item);
	if (data.isEmpty())
		return isegErrorProperty();

	IsegItemProperty result;
	memcpy(&result, data.constData(), sizeof(IsegItemProperty));
	return result;
}

quint32 IsegHalRemote::isegGetServerVersion()
{
	const QByteArrayList version = isegGetServerVersionString().split('.');
	if (version.size() != 4)
		return 0;

	return (version[0].toInt() << 24) + (version[1].toInt() << 16) +
			(version[2].toInt() << 8) + version[3].toInt();
}

QByteArray IsegHalRemote::isegGetServerVersionString()
{
	const QByteArray data = writeRead(ISEG_REMOTE_GETVERSION);

	return data;
}

IsegItem IsegHalRemote::isegErrorItem()
{
	IsegItem errorItem = EmptyIsegItem;
	strncpy(errorItem.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
	return errorItem;
}

IsegItemProperty IsegHalRemote::isegErrorProperty()
{
	IsegItemProperty errorProperty = EmptyIsegItemProperty;
	strncpy(errorProperty.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
	return errorProperty;
}

QByteArray IsegHalRemote::isegResult(IsegResult value)
{
	QByteArray result = QByteArray::number(value);

	switch (value) {
	case ISEG_OK:
		result.append(" (ISEG_OK)");
		break;
	case ISEG_ERROR:
		result.append(" (ISEG_ERROR)");
		break;
	case ISEG_WRONG_SESSION_NAME:
		result.append(" (ISEG_WRONG_SESSION_NAME)");
		break;
	case ISEG_WRONG_USER:
		result.append(" (ISEG_WRONG_USER)");
		break;
	case ISEG_WRONG_PASSWORD:
		result.append(" (ISEG_WRONG_PASSWORD)");
		break;
	case ISEG_NOT_AUTHORIZED:
		result.append(" (ISEG_NOT_AUTHORIZED)");
		break;
	case ISEG_NO_SSL_SUPPORT:
		result.append(" (ISEG_NO_SSL_SUPPORT)");
		break;
	default:
		result.append(" (UNKNOWN)");
		break;
	}

	return result;
}
