#include "iseghalremote.h"
#include "isegremoteapi.h"

#include <QDateTime>
#include <QHash>
#include <QMutex>
#include <QObject>
#include <QUuid>

#include <stdlib.h>

static const char NetworkTimeoutC[] = "NetworkTimeout";
static const char ServerVersionC[] = "ServerVersion";
static const char SessionNameC[] = "SessionName";

typedef QHash<QByteArray, IsegHalRemote *> IsegHalRemoteList;
typedef QHash<QByteArray, int> IsegHalRemoteTimeoutList;
static IsegHalRemoteList m_halRemoteList;
static IsegHalRemoteTimeoutList m_halTimeout;
static QMutex m_mutex;

static bool isKeyValid(const char *name)
{
	if (!m_halRemoteList.contains(name))
		return false;

	return true;
}

static QByteArray createTimeStamp()
{
	return QByteArray::number(QDateTime::currentDateTime().toTime_t()) + ".0000";
}

unsigned int iseg_getVersion(void)
{
	const unsigned int version[] = { VER_FILEVERSION };

	return (version[0] << 24) + (version[1] << 16) + (version[2] << 8) + version[3];
}

const char *iseg_getVersionString(void)
{
	return VER_FILEVERSION_STR;
}

IsegResult iseg_connect(const char *name, const char *interface, void *reserved)
{
	Q_UNUSED(reserved);

	if (isKeyValid(name))
		return ISEG_WRONG_SESSION_NAME;

	IsegHalRemote *hal = new IsegHalRemote();

	if (m_halTimeout.contains(name))
		hal->setTimeout(m_halTimeout[name]);

	const IsegResult result = hal->isegConnect(name, interface);

	if (result == ISEG_OK) {
		m_halRemoteList.insert(name, hal);
	} else {
		delete hal;
	}

	return result;
}

IsegResult iseg_disconnect(const char *name)
{
	if (!isKeyValid(name))
		return ISEG_WRONG_SESSION_NAME;

	IsegHalRemote *hal = m_halRemoteList.value(name);
	const IsegResult result = hal->isegDisconnect();

	delete hal;
	m_halRemoteList.remove(name);

	return result;
}

IsegResult iseg_setItem(const char *name, const char *object, const char *value)
{
	if (qstrncmp(object, NetworkTimeoutC, sizeof(NetworkTimeoutC)) == 0) {
		m_halTimeout[name] = 1000 * atoi(value);
		return ISEG_OK;
	}

	if (!isKeyValid(name))
		return ISEG_WRONG_SESSION_NAME;

	IsegHalRemote *hal = m_halRemoteList.value(name);
	return hal->isegSetItem(object, value);
}

static IsegItem getNetworkTimeout(int timeout)
{
	const QByteArray ts = createTimeStamp();

	IsegItem result = EmptyIsegItem;
	qstrncpy(result.object, NetworkTimeoutC, sizeof(result.object));
	qsnprintf(result.value, sizeof(result.value), "%d.%03d", timeout / 1000, timeout % 1000);
	qstrncpy(result.quality, ISEG_ITEM_QUALITY_OK, sizeof(result.quality));
	qstrncpy(result.timeStampLastChanged, ts.constData(), sizeof(result.timeStampLastChanged));
	qstrncpy(result.timeStampLastRefreshed, ts.constData(), sizeof(result.timeStampLastRefreshed));
	return result;
}

#if (QT_VERSION < QT_VERSION_CHECK(4, 8, 0))
#	define createUuidString() QUuid::createUuid().toString().toLatin1()
#else
#	define createUuidString() QUuid::createUuid().toByteArray()
#endif

static IsegItem getSessionName(const QByteArray &session, bool appendRandom)
{
	const QByteArray ts = createTimeStamp();

	QByteArray value = session;
	if (appendRandom) {
		const QByteArray id = createUuidString().mid(1, 8);
		value.append(id);
	}

	IsegItem result = EmptyIsegItem;
	qstrncpy(result.object, SessionNameC, sizeof(result.object));
	qstrncpy(result.value, value.constData(), sizeof(result.value));
	qstrncpy(result.quality, ISEG_ITEM_QUALITY_OK, sizeof(result.quality));
	qstrncpy(result.timeStampLastChanged, ts.constData(), sizeof(result.timeStampLastChanged));
	qstrncpy(result.timeStampLastRefreshed, ts.constData(), sizeof(result.timeStampLastRefreshed));
	return result;
}

static bool iseg_isInternalItem(const char *object)
{
	return (qstrncmp(object, NetworkTimeoutC, sizeof(NetworkTimeoutC)) == 0)
			|| (qstrncmp(object, SessionNameC, sizeof(SessionNameC)) == 0)
			|| (qstrncmp(object, ServerVersionC, sizeof(ServerVersionC)) == 0);
}

static IsegItem iseg_getInternalItem(IsegHalRemote *hal, const char *name, const char *object)
{
	IsegItem result = EmptyIsegItem;

	if (qstrncmp(object, NetworkTimeoutC, sizeof(NetworkTimeoutC)) == 0) {
		if (hal)
			return getNetworkTimeout(hal->timeout());

		if (m_halTimeout.contains(name))
			return getNetworkTimeout(m_halTimeout.value(name));

		return getNetworkTimeout(IsegHalRemote::DefaultNetworkTimeout);
	}

	if (qstrncmp(object, SessionNameC, sizeof(SessionNameC)) == 0) {
		return getSessionName(name, hal == NULL);
	}

	if (qstrncmp(object, ServerVersionC, sizeof(ServerVersionC)) == 0) {
		const QByteArray ts = createTimeStamp();
		qstrncpy(result.object, object, sizeof(result.object));
		qstrncpy(result.value, hal->isegGetServerVersionString().constData(), sizeof(result.value));
		qstrncpy(result.quality, ISEG_ITEM_QUALITY_OK, sizeof(result.quality));
		qstrncpy(result.timeStampLastChanged, ts.constData(), sizeof(result.timeStampLastChanged));
		qstrncpy(result.timeStampLastRefreshed, ts.constData(), sizeof(result.timeStampLastRefreshed));
	}

	return result;
}

IsegItem iseg_getItem(const char *name, const char *object)
{
	if (!isKeyValid(name)) {
		if (iseg_isInternalItem(object))
			return iseg_getInternalItem(NULL, name, object);

		return IsegHalRemote::isegErrorItem();
	}

	IsegHalRemote *hal = m_halRemoteList.value(name);

	if (iseg_isInternalItem(object))
		return iseg_getInternalItem(hal, name, object);

	return hal->isegGetItem(object);
}

/**
 * @brief Returns an IsegItem formated as semicolon-separated string.
 *
 * This function works like iseg_getItem(), but formats the result as single
 * string and separates the fields with semicolons. This is useful for LabVIEW,
 * which does not support structures as return value.
 *
 * Example:
 *  char buf[512];
 *  IsegResult result = iseg_getItemString("can0", "0.Status", buf, sizeof(buf));
 *  // buf = "0.Status;00;002;1452786303.6296;1452786303.6159"
 *
 * @param name   The session name used for iseg_connect()
 * @param object The item that should be read, e.g. "0.ModuleList"
 * @param result The buffer to fill with the resultating string
 * @param size   The size of the given buffer
 * @return ISEG_OK on success, otherwise an error code
 */
IsegResult iseg_getItemString(const char *name, const char *object, char *result, int size)
{
	const IsegItem item = iseg_getItem(name, object);

	qsnprintf(result, size, "%s;%s;%s;%s;%s", item.object, item.value,
			  item.quality, item.timeStampLastRefreshed, item.timeStampLastChanged);

	return ISEG_OK;
}

static IsegItemProperty iseg_getInternalItemProperty(const char *object)
{
	IsegItemProperty result = EmptyIsegItemProperty;

	if (qstrncmp(object, NetworkTimeoutC, sizeof(NetworkTimeoutC)) == 0) {
		qstrncpy(result.object, object, sizeof(result.object));
		qstrncpy(result.type, "R4", sizeof(result.type));
		qstrncpy(result.unit, "s", sizeof(result.unit));
		qstrncpy(result.access, "RW", sizeof(result.access));
		qstrncpy(result.quality, ISEG_ITEM_QUALITY_OK, sizeof(result.quality));
	} else if (qstrncmp(object, SessionNameC, sizeof(SessionNameC)) == 0) {
		qstrncpy(result.object, object, sizeof(result.object));
		qstrncpy(result.type, "STR", sizeof(result.type));
		qstrncpy(result.access, "R", sizeof(result.access));
		qstrncpy(result.quality, ISEG_ITEM_QUALITY_OK, sizeof(result.quality));
	} else if (qstrncmp(object, ServerVersionC, sizeof(ServerVersionC)) == 0) {
		qstrncpy(result.object, object, sizeof(result.object));
		qstrncpy(result.type, "STR", sizeof(result.type));
		qstrncpy(result.access, "R", sizeof(result.access));
		qstrncpy(result.quality, ISEG_ITEM_QUALITY_OK, sizeof(result.quality));
	}

	return result;
}

IsegItemProperty iseg_getItemProperty(const char *name, const char *object)
{
	if (!isKeyValid(name))
		return IsegHalRemote::isegErrorProperty();

	IsegHalRemote *hal = m_halRemoteList.value(name);

	if (iseg_isInternalItem(object))
		return iseg_getInternalItemProperty(object);

	return hal->isegGetItemProperty(object);
}

/**
 * @brief Returns an IsegItemProperty formated as semicolon-separated string.
 *
 * This function works like iseg_getItemProperty(), but formats the result as
 * single string and separates the fields with semicolons. This is useful for
 * LabVIEW, which does not support structures as return value.
 *
 * Example:
 *  char buf[512];
 *  IsegResult result = iseg_getItemPropertyString("can0", "0.Status", buf, sizeof(buf));
 *  // buf = "0.Status;STR;;R;002"
 *
 * @param name   The session name used for iseg_connect()
 * @param object The item that should be read, e.g. "0.ModuleList"
 * @param result The buffer to fill with the resultating string
 * @param size   The size of the given buffer
 * @return ISEG_OK on success, otherwise an error code
 */
IsegResult iseg_getItemPropertyString(const char *name, const char *object, char *result, int size)
{
	const IsegItemProperty property = iseg_getItemProperty(name, object);

	qsnprintf(result, size, "%s;%s;%s;%s;%s", property.object, property.type,
			  property.unit, property.access, property.quality);

	return ISEG_OK;
}
