#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "isegremoteapi.h"
#include "iseghalremote.h"

#include <QCloseEvent>
#include <QCompleter>
#include <QFileDialog>
#include <QSettings>
#include <QTimer>

static const char HostsC[]    = "Hosts";
static const char PortsC[]    = "Ports";
static const char CommandsC[] = "Commands";

static QString isegResult(IsegResult result)
{
	return QLatin1String(IsegHalRemote::isegResult(result));
}

QStringList MainWindow::getSettingValues(const QString &section)
{
	QStringList result;
	m_settings->beginGroup(section);
	const QStringList keys = m_settings->allKeys();
	foreach (const QString &key, keys)
		result.append(m_settings->value(key).toString());
	m_settings->endGroup();

	return result;
}

void MainWindow::appendSettingValue(const QString &section, const QString &value)
{
	m_settings->beginGroup(section);

	QStringList values = QStringList() << value;
	const QStringList keys = m_settings->allKeys();
	for (int i = 0; i < keys.size(); ++i) {
		const QString address = m_settings->value(keys[i]).toString();
		if (!values.contains(address))
			values.append(address);
		if (values.size() > 10)
			break;
	}

	m_settings->remove(QString());

	for (int i = 0; i < values.size(); ++i)
		m_settings->setValue(QString::number(i), values[i]);

	m_settings->endGroup();
}

void MainWindow::setSettingValues(const QString &section, const QStringList &values)
{
	m_settings->beginGroup(section);

	m_settings->remove(QString());

	for (int i = 0, size = values.size(); i < size; ++i)
		m_settings->setValue(QString::number(i), values[i]);

	m_settings->endGroup();
}

MainWindow::MainWindow(QWidget *parent) :
	QMainWindow(parent),
	m_connected(false),
	m_timer(0),
	ui(new Ui::MainWindow)
{
	ui->setupUi(this);

	ui->groupCommand->setEnabled(false);
	ui->groupScript->setEnabled(false);

	setWindowTitle(tr("iseg HAL Remote Example %1").arg(VER_FILEVERSION_STR));

	ui->command->completer()->setCaseSensitivity(Qt::CaseSensitive);
	connect(ui->command, SIGNAL(currentIndexChanged(QString)),
			this, SLOT(currentCommandChanged(QString)));
	connect(ui->command->lineEdit(), SIGNAL(returnPressed()),
			this, SLOT(on_buttonSend_clicked()));

	ui->hint->setText(tr(
		"Hints:"
		"<ul>"
			"<li>To get a value: 0.0.1.VoltageSet?</li>"
			"<li>To set a value: 0.0.1.VoltageSet=1234</li>"
			"<li>To get the properties: 0.0.1.VoltageSet</li>"
		"</ul>"));

	m_settings = new QSettings(QSettings::IniFormat, QSettings::UserScope,
							   QLatin1String("iseg Spezialelektronik GmbH"),
							   QLatin1String("isegHalRemoteExample"));

	const QStringList hosts = getSettingValues(QLatin1String(HostsC));
	if (!hosts.isEmpty()) {
		ui->host->clear();
		ui->host->addItems(hosts);
	}

	const QStringList ports = getSettingValues(QLatin1String(PortsC));
	if (!ports.isEmpty()) {
		ui->port->clear();
		ui->port->addItems(ports);
	}

	const QStringList commands = getSettingValues(QLatin1String(CommandsC));
	if (!commands.isEmpty()) {
		ui->command->clear();
		ui->command->addItems(commands);
	}
}

MainWindow::~MainWindow()
{
	delete ui;
}

void MainWindow::closeEvent(QCloseEvent *event)
{
	if (m_connected)
		iseg_disconnect(m_session.constData());

	QStringList list;
	const int count = ui->command->count();
	for (int i = qMax(count - 50, 0); i < count; ++i)
		list.append(ui->command->itemText(i));

	setSettingValues(QLatin1String(CommandsC), list);

	event->accept();
}

static QString isegQuality(const QString &quality)
{
	QString result = quality;

	if (quality == QLatin1String(ISEG_ITEM_QUALITY_INVALID)) {
		result.append(QLatin1String(" (ISEG_ITEM_QUALITY_INVALID)"));
	} else if (quality == QLatin1String(ISEG_ITEM_QUALITY_INITIALIZE)) {
		result.append(QLatin1String(" (ISEG_ITEM_QUALITY_INITIALIZE)"));
	} else if (quality == QLatin1String(ISEG_ITEM_QUALITY_OK)) {
		result.append(QLatin1String(" (ISEG_ITEM_QUALITY_OK)"));
	} else if (quality == QLatin1String(ISEG_ITEM_QUALITY_COMMUNICATION_BAD)) {
		result.append(QLatin1String(" (ISEG_ITEM_QUALITY_COMMUNICATION_BAD)"));
	} else {
		result.append(QLatin1String(" (ISEG_ITEM_QUALITY_ERROR)"));
	}

	return result;
}

void MainWindow::on_login_clicked()
{
	const QString interface = QString::fromLatin1("hal://%1:%2/%3,%4,%5")
			.arg(ui->host->currentText())
			.arg(ui->port->currentText())
			.arg(ui->canInterface->currentText())
			.arg(ui->user->text())
			.arg(ui->password->text());

	const IsegItem session = iseg_getItem("myEth", "SessionName");
	m_session = session.value;

	const QByteArray timeout = QByteArray::number(1E-3 * ui->timeout->value());
	IsegResult result = iseg_setItem(m_session.constData(), "NetworkTimeout", timeout.constData());
	ui->log->append(tr("Set timeout returned: %1").arg(isegResult(result)));

	result = iseg_connect(m_session.constData(), interface.toLatin1().constData(), NULL);
	ui->log->append(tr("Login returned: %1").arg(isegResult(result)));

	if (result == ISEG_NO_SSL_SUPPORT) {
		ui->log->setTextColor(Qt::red);
		ui->log->append(tr("No SSL support detected."));
		ui->log->append(tr("Install the OpenSSL libraries and try again."));
		ui->log->setTextColor(Qt::black);
	}

	if (result != ISEG_OK)
		return;

	const IsegItem version = iseg_getItem(m_session.constData(), "ServerVersion");
	ui->log->append(tr("Server Version: %1").arg(version.value));

	appendSettingValue(QLatin1String(HostsC), ui->host->currentText());
	appendSettingValue(QLatin1String(PortsC), ui->port->currentText());

	m_connected = true;
	ui->groupLogin->setEnabled(false);
	ui->groupCommand->setEnabled(true);
	ui->groupScript->setEnabled(true);
	setFocusToCommand();
}

void MainWindow::setFocusToCommand()
{
	ui->command->setFocus();
	ui->command->lineEdit()->selectAll();
}

QStringList MainWindow::sendCommand(const QString &command)
{
	QStringList result;
	result.append(QLatin1String("---------------------------------------"));

	const QStringList data = command.trimmed().split(QLatin1Char('='));
	QString nextCommand = data[0];
	if (data.size() == 1) {
		if (nextCommand.endsWith(QLatin1Char('?'))) {
			nextCommand.chop(1);
			const IsegItem item = iseg_getItem(m_session.constData(), nextCommand.toLatin1());

			result.append(tr("Object:    %1").arg(item.object));
			result.append(tr("Value:     %1").arg(item.value));
			result.append(tr("Quality:   %1").arg(isegQuality(item.quality)));
			result.append(tr("Changed:   %1").arg(item.timeStampLastRefreshed));
			result.append(tr("Refreshed: %1").arg(item.timeStampLastChanged));
		} else {
			const IsegItemProperty prop = iseg_getItemProperty(m_session.constData(), nextCommand.toLatin1());
			result.append(tr("Object:    %1").arg(prop.object));
			result.append(tr("Type:      %1").arg(prop.type));
			result.append(tr("Quality:   %1").arg(isegQuality(prop.quality)));
			result.append(tr("Unit:      %1").arg(prop.unit));
			result.append(tr("Access:    %1").arg(prop.access));
		}
	} else {
		IsegResult set = iseg_setItem(m_session.constData(), nextCommand.toLatin1(), data[1].toLatin1());

		result.append(tr("%1 returned %2")
			.arg(ui->command->currentText()).arg(isegResult(set)));
	}

	return result;
}

void MainWindow::on_buttonSend_clicked()
{
	const QStringList lines = sendCommand(ui->command->currentText());
	foreach (const QString &line, lines)
		ui->log->append(line);

	setFocusToCommand();
}

void MainWindow::currentCommandChanged(const QString &newText)
{
	// Workaround for QTBUG-41288
	if (ui->command->lineEdit() && ui->command->lineEdit()->completer())
		ui->command->lineEdit()->completer()->setCompletionPrefix(newText);
}

void MainWindow::on_timeout_valueChanged(int arg1)
{
	Q_UNUSED(arg1);
	// TODO Define iseg API function for timeout setting
}

void MainWindow::on_selectScript_clicked()
{
	const QString workingDirectory = QDir::fromNativeSeparators(ui->scriptFilename->text());
	const QString filter = tr("Script files (*.txt)");
	const QString fileName = QFileDialog::getOpenFileName(this,
		tr("Select script file"), workingDirectory, filter);

	if (fileName.isEmpty())
		return;

	ui->scriptFilename->setText(QDir::toNativeSeparators(fileName));}

void MainWindow::on_runScript_clicked()
{
	if (m_timer) {
		ui->runScript->setText(tr("Run"));
		delete m_timer;
		m_timer = 0;
	} else {
		QFile scriptFile(QDir::fromNativeSeparators(ui->scriptFilename->text()));
		scriptFile.open(QIODevice::ReadOnly);
		m_scriptLines = scriptFile.readAll().split('\n');
		m_scriptLines.removeAll(QByteArray());
		m_currentScriptLine = 0;
		scriptFile.close();

		ui->runScript->setText(tr("Stop"));
		m_timer = new QTimer(this);
		m_timer->singleShot(50, this, SLOT(sendNextScriptLine()));
	}
}

void MainWindow::sendNextScriptLine()
{
	if (m_currentScriptLine >= m_scriptLines.size())
		m_currentScriptLine = 0;
	const QByteArray &command = m_scriptLines[m_currentScriptLine++].trimmed();

	if (command == "STOP") {
		ui->runScript->setText(tr("Run"));
		delete m_timer;
		m_timer = 0;
		ui->log->append(tr("Stopping script."));
		return;
	}

	if (command.startsWith("SLEEP ")) {
		const int time = command.mid(6).trimmed().toInt();
		m_timer->singleShot(time, this, SLOT(sendNextScriptLine()));
		ui->log->append(tr("Sleeping for %1 ms...").arg(time));
		return;
	}

	const QStringList lines = sendCommand(command);
	foreach (const QString &line, lines)
		ui->log->append(line);

	m_timer->singleShot(50, this, SLOT(sendNextScriptLine()));
}
