#include "../isegapi.h"
#include "iseghalremotecommon.h"
#include "iseghalremoteconnection.h"

#include <QFile>
#include <QHostAddress>
#include <QSslSocket>
#include <QSslKey>
#include <QSslConfiguration>

IsegHalRemoteConnection::IsegHalRemoteConnection(quint16 socketDescriptor, QObject *parent) :
	QObject(parent),
	m_authorized(false)
{
	m_socket = new QSslSocket(this);

	if (!m_socket->setSocketDescriptor(socketDescriptor)) {
		delete m_socket;
		return;
	}

	QFile certFile(QLatin1String(":/localhost.cert"));
	QFile keyFile(QLatin1String(":/localhost.key"));
	certFile.open(QIODevice::ReadOnly);
	keyFile.open(QIODevice::ReadOnly);
	QSslCertificate certificate(&certFile, QSsl::Pem);
	QSslKey sslKey(&keyFile, QSsl::Rsa, QSsl::Pem);
	certFile.close();
	keyFile.close();

	QSslConfiguration sslConfiguration;
	sslConfiguration.setPeerVerifyMode(QSslSocket::VerifyNone);
	sslConfiguration.setLocalCertificate(certificate);
	sslConfiguration.setPrivateKey(sslKey);
	sslConfiguration.setProtocol(QSecureSslProtocols);
	m_socket->setSslConfiguration(sslConfiguration);
	m_socket->startServerEncryption();

	// Connect the SSL socket's signals to our slots.
	connect(m_socket, SIGNAL(encrypted()), this, SLOT(ready()));
	connect(m_socket, SIGNAL(disconnected()), SLOT(connectionClosed()));
	connect(m_socket, SIGNAL(readyRead()), SLOT(readData()));
	connect(m_socket, SIGNAL(error(QAbstractSocket::SocketError)),
			SLOT(error(QAbstractSocket::SocketError)));
}

IsegHalRemoteConnection::~IsegHalRemoteConnection()
{
	// Report that the connection has closed.
	qDebug("Connection closed.");
}

void IsegHalRemoteConnection::ready()
{
	// Provide feedback to the user about incoming connections. This
	// slot is only called if the connection was established, so all
	// communication is now encrypted.
	qDebug("Accepted new client from %s:%d",
		   qPrintable(m_socket->peerAddress().toString()),
		   m_socket->peerPort());
}

static QByteArray buildResult(char type, const QByteArray &session, const QByteArray &data)
{
	return type + session + ";" + data;
}

QByteArray IsegHalRemoteConnection::isegConnect(const QByteArray &data)
{
	// cMYETH;hal://x.y.z.y:port/can0,user,pass
	const QByteArrayList list = data.mid(1).split(';');
	Q_ASSERT(list.size() >= 2);

	const QByteArrayList interface = list[1].mid(list[1].lastIndexOf('/') + 1).split(',');
	Q_ASSERT(interface.size() >= 3);

	// TODO user,passw check
	if (interface[1] != "user")
		return buildResult(ISEG_REMOTE_CONNECT, list[0], QByteArray::number(ISEG_WRONG_USER));

	if (interface[2] != "pass")
		return buildResult(ISEG_REMOTE_CONNECT, list[0], QByteArray::number(ISEG_WRONG_PASSWORD));

	m_session.clear();
	m_authorized = true;

	const IsegResult result = iseg_connect(list[0].constData(), interface[0].constData(), NULL);
	if (result == ISEG_OK)
		m_session = list[0];

	return buildResult(ISEG_REMOTE_CONNECT, list[0], QByteArray::number(result));
}

QByteArray IsegHalRemoteConnection::isegDisconnect(const QByteArray &data)
{
	const QByteArray payload = data.mid(1);
	m_session.clear();
	m_authorized = false;

	const IsegResult result = iseg_disconnect(payload.constData());
	return buildResult(ISEG_REMOTE_DISCONNECT, payload, QByteArray::number(result));
}

QByteArray IsegHalRemoteConnection::isegGetItem(const QByteArray &data)
{
	const QByteArrayList list = data.mid(1).split(';');
	Q_ASSERT(list.size() == 2);

	if (!m_authorized)
		return buildResult(ISEG_REMOTE_GETITEM, list[0], QByteArray::number(ISEG_NOT_AUTHORIZED));

	const IsegItem item = iseg_getItem(list[0].constData(), list[1].constData());

	return buildResult(ISEG_REMOTE_GETITEM, list[0],
			QByteArray::fromRawData((const char *)&item, sizeof(item)));
}

QByteArray IsegHalRemoteConnection::isegSetItem(const QByteArray &data)
{
	const QByteArrayList list = data.mid(1).split(';');
	Q_ASSERT(list.size() == 2);

	if (!m_authorized)
		return buildResult(ISEG_REMOTE_SETITEM, list[0], QByteArray::number(ISEG_NOT_AUTHORIZED));

	const QByteArrayList list2 = list[1].split(',');

	const IsegResult result =
			iseg_setItem(list[0].constData(), list2[0].constData(), list2[1].constData());

	return buildResult(ISEG_REMOTE_SETITEM, list[0], QByteArray::number(result));
}

QByteArray IsegHalRemoteConnection::isegGetProperty(const QByteArray &data)
{
	const QByteArrayList list = data.mid(1).split(';');
	Q_ASSERT(list.size() == 2);

	if (!m_authorized)
		return buildResult(ISEG_REMOTE_GETPROPERTY, list[0], QByteArray::number(ISEG_NOT_AUTHORIZED));

	const IsegItemProperty prop = iseg_getItemProperty(list[0].constData(), list[1].constData());

	return buildResult(ISEG_REMOTE_GETPROPERTY, list[0],
			QByteArray::fromRawData((const char *)&prop, sizeof(prop)));
}

QByteArray IsegHalRemoteConnection::isegGetVersion(const QByteArray &data)
{
	return buildResult(ISEG_REMOTE_GETVERSION, data.mid(1), QByteArray(VER_FILEVERSION_STR));
}

void IsegHalRemoteConnection::readData()
{
	const QByteArray data = m_socket->readAll();

	if (data.size() < 1)
		return;

	switch (data[0]) {
	case ISEG_REMOTE_CONNECT:
		m_socket->write(isegConnect(data));
		break;

	case ISEG_REMOTE_DISCONNECT:
		m_socket->write(isegDisconnect(data));
		break;

	case ISEG_REMOTE_GETITEM:
		m_socket->write(isegGetItem(data));
		break;

	case ISEG_REMOTE_SETITEM:
		m_socket->write(isegSetItem(data));
		break;

	case ISEG_REMOTE_GETPROPERTY:
		m_socket->write(isegGetProperty(data));
		break;

	case ISEG_REMOTE_GETVERSION:
		m_socket->write(isegGetVersion(data));
		break;

	default:
		// TODO
		break;
	}
}

void IsegHalRemoteConnection::connectionClosed()
{
	if (!m_session.isEmpty()) {
		qDebug("Socket closing while still connected to isegHAL... disconnecting.");
		iseg_disconnect(m_session);
		m_session.clear();
	}

	// Although the socket may be closing, we must not delete it until
	// the delayed close is done.
	if (m_socket->state() == QAbstractSocket::ClosingState) {
		connect(m_socket, SIGNAL(disconnected()), SLOT(deleteLater()));
	} else {
		deleteLater();
		return;
	}

	qDebug("Connection closed.");
}

void IsegHalRemoteConnection::error(QAbstractSocket::SocketError)
{
	// The SSL socket conveniently provides human readable error
	// messages through the errorString() call. Note that sometimes
	// the errors come directly from the underlying SSL library, and
	// the quality of the text may vary.
	qDebug("Error: %s", qPrintable(m_socket->errorString()));
}
