#include "clientthread.h"
#include "isegremoteapi.h"
#include "iseghalremote.h"
#include "iseghalremotecommon.h"
#include "kcanapi.h"
#include "sessionthread.h"

#include <QString>
#include <QProcess>
#include <QtTest>
#include <QSet>
#include <QSslSocket>

static const char ServerNameC[]    = "isegHalRemoteServer-direct";
static const char SessionC[]       = "IsegHalRemoteTest";
static const char SecondSessionC[] = "IsegHalTest2";
static const char WrongHostC[]     = "123.1.2.3";
static const char InterfaceC[]     = "can0";
static const char UserNameC[]      = "user";
static const char PassWordC[]      = "pass";
static const char WrongUserNameC[] = "iseg";
static const char WrongPassWordC[] = "iseg";

static const char NetworkTimeoutC[] = "NetworkTimeout";
static const char NetworkTimeoutDefaultValueC[] = "5.000";
static const char NetworkTimeoutValueC[] = "2.000";
static const char ServerVersionC[] = "ServerVersion";
static const char SessionNameC[] = "SessionName";

namespace QTest {

template<>
char *toString(const IsegResult &result)
{
	QByteArray ba = IsegHalRemote::isegResult(result);

	return qstrdup(ba.data());
}

} // namespace QTest

#define _(x) QByteArray(x)

class IsegHalRemoteTest : public QObject
{
	Q_OBJECT

public:
	IsegHalRemoteTest();

	void setHost(const QByteArray &host);

private slots:
	void initTestCase();
	void cleanupTestCase();

	void server_isRunning();

	void networkTimeout_getDefault();
	void networkTimeout_setBeforeConnect();
	void networkTimeout_getBeforeConnect();
	void sessionName_getBeforeConnect();
	void sessionName_isUnique();

	void isegConnect_wrongHost();
	void isegConnect_wrongPort();
	void isegConnect_wrongUser();
	void isegConnect_wrongPassword();
	void isegConnect_ok();
	void isegConnect_secondClient();
	void isegConnect_alreadyConnected();

	void getItem_serverVersion();
	void getProperty_serverVersion();
	void getItem_networkTimeout();
	void getProperty_networkTimeout();
	void getItem_sessionName();
	void getProperty_sessionName();

	void getItem_bitRate();
	void getItem_canBusStatus();
	void getItemString_canBusStatus();

	void getItem_benchmark();

	void getProperty();
	void getPropertyString();

	void checkServerStability();

	void getItem_parallel();

	void isegDisconnect_ok();
	void isegDisconnect_secondClient();
	void isegDisconnect_notConnected();

private:
	QByteArray connectString(const QByteArray &host, quint16 port,
		const QByteArray &interface, const QByteArray &userName, const QByteArray &passWord);

	QProcess m_process;

	QByteArray m_host;
	bool m_local;
};

IsegHalRemoteTest::IsegHalRemoteTest() :
	m_host("127.0.0.1"),
	m_local(true)
{
}

void IsegHalRemoteTest::initTestCase()
{
	if (!m_local)
		return;

	const QString serverName = QString::fromLatin1("../%1/%1").arg(ServerNameC);
	const QString linkerPath = QLatin1String("LD_LIBRARY_PATH=..");
	QStringList env = m_process.environment();
	env.append(linkerPath);

	m_process.setEnvironment(env);
	m_process.setProcessChannelMode(QProcess::MergedChannels);
	m_process.start(serverName, QStringList(), QIODevice::ReadOnly);
	QVERIFY(m_process.waitForStarted(10000));
}

void IsegHalRemoteTest::cleanupTestCase()
{
	if (!m_local)
		return;

	m_process.kill();
	QVERIFY(m_process.waitForFinished(5000));
}

void IsegHalRemoteTest::server_isRunning()
{
	if (!m_local)
		return;

	m_process.waitForReadyRead(5000);
	const QByteArray read = m_process.readAll();
	const QByteArray expected = _("Listening on port ")
			+ QByteArray::number(ISEG_REMOTE_PORT) + _(". Press Ctrl-C to exit.\n");

	QCOMPARE(read, expected);
}

void IsegHalRemoteTest::networkTimeout_getDefault()
{
	const IsegItem result = iseg_getItem(SessionC, NetworkTimeoutC);

	QCOMPARE(result.object,  NetworkTimeoutC);
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.value,   NetworkTimeoutDefaultValueC);
}

void IsegHalRemoteTest::networkTimeout_setBeforeConnect()
{
	const IsegResult setResult = iseg_setItem(SessionC, NetworkTimeoutC, NetworkTimeoutValueC);

	QCOMPARE(setResult, ISEG_OK);
}

void IsegHalRemoteTest::networkTimeout_getBeforeConnect()
{
	const IsegItem result = iseg_getItem(SessionC, NetworkTimeoutC);

	QCOMPARE(result.object,  NetworkTimeoutC);
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.value,   NetworkTimeoutValueC);
}

void IsegHalRemoteTest::sessionName_getBeforeConnect()
{
	const IsegItem result = iseg_getItem(SessionC, SessionNameC);

	QCOMPARE(result.object,  SessionNameC);
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);

	const QByteArray value = result.value;
	QVERIFY(value.startsWith(SessionC));
	QCOMPARE(value.size(), int(strlen(SessionC) + 8));
	// "IsegHalRemoteTestc89918d7"
}

void IsegHalRemoteTest::sessionName_isUnique()
{
	enum { ThreadNumber = 10 };
	SessionThread thread[ThreadNumber];
	QSet<QByteArray> set;

	for (int i = 0; i < ThreadNumber; ++i) {
		thread[i].start();
	}

	for (int i = 0; i < ThreadNumber; ++i) {
		thread[i].wait(100);
		const QByteArray name = thread[i].sessionName();
		set.insert(name);

		if (set.size() != i + 1) {
			qDebug() << "set = " << set;
			qDebug() << "name = " << name;
		}
	}

	QCOMPARE(set.size(), int(ThreadNumber));
}

void IsegHalRemoteTest::isegConnect_wrongHost()
{
	const QByteArray interface = connectString(WrongHostC, ISEG_REMOTE_PORT, InterfaceC, UserNameC, PassWordC);

	const IsegResult result = iseg_connect(SessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_ERROR);
}

void IsegHalRemoteTest::isegConnect_wrongPort()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT + 1, InterfaceC, UserNameC, PassWordC);

	const IsegResult result = iseg_connect(SessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_ERROR);
}

void IsegHalRemoteTest::isegConnect_wrongUser()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT, InterfaceC, WrongUserNameC, PassWordC);

	const IsegResult result = iseg_connect(SessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_WRONG_USER);
}

void IsegHalRemoteTest::isegConnect_wrongPassword()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT, InterfaceC, UserNameC, WrongPassWordC);

	const IsegResult result = iseg_connect(SessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_WRONG_PASSWORD);
}

void IsegHalRemoteTest::isegConnect_ok()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT, InterfaceC, UserNameC, PassWordC);

	const IsegResult result = iseg_connect(SessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_OK);
}

void IsegHalRemoteTest::isegConnect_secondClient()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT, InterfaceC, UserNameC, PassWordC);

	const IsegResult result = iseg_connect(SecondSessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_OK);
}

void IsegHalRemoteTest::isegConnect_alreadyConnected()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT, InterfaceC, UserNameC, PassWordC);

	const IsegResult result = iseg_connect(SessionC, interface.constData(), NULL);

	QCOMPARE(result, ISEG_WRONG_SESSION_NAME);
}

void IsegHalRemoteTest::getItem_serverVersion()
{
	const IsegItem result = iseg_getItem(SessionC, ServerVersionC);

	QCOMPARE(result.object,  ServerVersionC);
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.value,   VER_FILEVERSION_STR);
}

void IsegHalRemoteTest::getProperty_serverVersion()
{
	const IsegItemProperty result = iseg_getItemProperty(SessionC, ServerVersionC);

	QCOMPARE(result.object,  ServerVersionC);
	QCOMPARE(result.type,    "STR");
	QCOMPARE(result.unit,    "");
	QCOMPARE(result.access,  "R");
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
}

void IsegHalRemoteTest::getItem_networkTimeout()
{
	const IsegItem result = iseg_getItem(SessionC, NetworkTimeoutC);

	QCOMPARE(result.object,  NetworkTimeoutC);
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.value,   NetworkTimeoutValueC);
}

void IsegHalRemoteTest::getProperty_networkTimeout()
{
	const IsegItemProperty result = iseg_getItemProperty(SessionC, NetworkTimeoutC);

	QCOMPARE(result.object,  NetworkTimeoutC);
	QCOMPARE(result.type,    "R4");
	QCOMPARE(result.unit,    "s");
	QCOMPARE(result.access,  "RW");
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
}

void IsegHalRemoteTest::getItem_sessionName()
{
	const IsegItem result = iseg_getItem(SessionC, SessionNameC);

	QCOMPARE(result.object,  SessionNameC);
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.value,   SessionC);
}

void IsegHalRemoteTest::getProperty_sessionName()
{
	const IsegItemProperty result = iseg_getItemProperty(SessionC, SessionNameC);

	QCOMPARE(result.object,  SessionNameC);
	QCOMPARE(result.type,    "STR");
	QCOMPARE(result.unit,    "");
	QCOMPARE(result.access,  "R");
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
}

void IsegHalRemoteTest::getItem_bitRate()
{
	const IsegItem result = iseg_getItem(SessionC, "0.BitRate");

	const int value = atoi(result.value) / 1000;
	QCOMPARE(result.object, "0.BitRate");
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QVERIFY(value == 125 || value == 250 || value == 1000);
}

void IsegHalRemoteTest::getItem_canBusStatus()
{
	const IsegItem result = iseg_getItem(SessionC, "0.Status");

	QCOMPARE(result.object, "0.Status");
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.value, CAN_STATE_ERROR_ACTIVE_STRING);
}

void IsegHalRemoteTest::getItemString_canBusStatus()
{
	char buf[512];

	const IsegResult result = iseg_getItemString(SessionC, "0.Status", buf, sizeof(buf));

	const QByteArrayList list = QByteArray(buf).split(';');
	QCOMPARE(result, ISEG_OK);
	QCOMPARE(list.size(), 5);
	QCOMPARE(list[0], _("0.Status"));                    // object
	QCOMPARE(list[1], _(CAN_STATE_ERROR_ACTIVE_STRING)); // value
	QCOMPARE(list[2], _(ISEG_ITEM_QUALITY_OK));          // quality
}

void IsegHalRemoteTest::getItem_benchmark()
{
	QBENCHMARK {
		iseg_getItem(SessionC, "0.BitRate");
	}
}

void IsegHalRemoteTest::getProperty()
{
	const IsegItemProperty result = iseg_getItemProperty(SessionC, "0.Status");

	QCOMPARE(result.object, "0.Status");
	QCOMPARE(result.access, "R");
	QCOMPARE(result.quality, ISEG_ITEM_QUALITY_OK);
	QCOMPARE(result.type, "STR");
	QCOMPARE(result.unit, "");
}

void IsegHalRemoteTest::getPropertyString()
{
	char buf[512];

	const IsegResult result = iseg_getItemPropertyString(SessionC, "0.Status", buf, sizeof(buf));

	const QByteArrayList list = QByteArray(buf).split(';');
	QCOMPARE(result, ISEG_OK);
	QCOMPARE(list.size(), 5);
	QCOMPARE(list[0], _("0.Status"));           // object
	QCOMPARE(list[1], _("STR"));                // type
	QCOMPARE(list[2], _(""));                   // unit
	QCOMPARE(list[3], _("R"));                  // access
	QCOMPARE(list[4], _(ISEG_ITEM_QUALITY_OK)); // quality
}

void IsegHalRemoteTest::checkServerStability()
{
	QByteArray data;
	data.resize(50000);
	for (int i = 0; i < data.size(); ++i)
		data[i] = 'A' + qrand() % ('Z' - 'A');

	QSslSocket socket;
	socket.setProtocol(QSecureSslProtocols);
	socket.setPeerVerifyMode(QSslSocket::VerifyNone);
	socket.connectToHostEncrypted(m_host, ISEG_REMOTE_PORT);
	QVERIFY(socket.waitForEncrypted());

	socket.write(data);
	QVERIFY(socket.waitForBytesWritten(1000));

	if (!m_local)
		return;

	m_process.waitForReadyRead(1000);
	QByteArray output = m_process.readAll();
	QVERIFY(!output.isEmpty());
}

void IsegHalRemoteTest::getItem_parallel()
{
	const QByteArray interface = connectString(m_host, ISEG_REMOTE_PORT, InterfaceC, UserNameC, PassWordC);

	ClientThread statusThread("theFirst", interface, "0.Status");
	ClientThread bitRateThread("theSecond", interface, "0.BitRate");
	statusThread.start();
	bitRateThread.start();

	QVERIFY(statusThread.wait(5000));
	QVERIFY(bitRateThread.wait(5000));
	QVERIFY(statusThread.success());
	QVERIFY(bitRateThread.success());
}

void IsegHalRemoteTest::isegDisconnect_ok()
{
	const IsegResult result = iseg_disconnect(SessionC);

	QCOMPARE(result, ISEG_OK);
}

void IsegHalRemoteTest::isegDisconnect_secondClient()
{
	const IsegResult result = iseg_disconnect(SecondSessionC);

	QCOMPARE(result, ISEG_OK);
}

void IsegHalRemoteTest::isegDisconnect_notConnected()
{
	const IsegResult result = iseg_disconnect(SessionC);

	QCOMPARE(result, ISEG_WRONG_SESSION_NAME);
}

QByteArray IsegHalRemoteTest::connectString(const QByteArray &host, quint16 port,
											const QByteArray &interface,
											const QByteArray &userName, const QByteArray &passWord)
{
	const QByteArray result = _("hal://") + host + _(":") + QByteArray::number(port)
			+ _("/") + interface + _(",") + userName + _(",") + passWord;

	return result;
}

void IsegHalRemoteTest::setHost(const QByteArray &host)
{
	qDebug() << "Testing against remote host " << host;
	m_host = host;
	m_local = false;
}

int main(int argc, char *argv[])
{
	QStringList params;
	IsegHalRemoteTest test;

	for (int i = 0; i < argc; ++i)
		params << QLatin1String(argv[i]);

	if ((params.count() > 2) && (params[1] == QLatin1String("--ip"))) {
		test.setHost(params[2].toLatin1());
		params.removeAt(2);
		params.removeAt(1);
	}

	return QTest::qExec(&test, params);
}

#include "iseghalremotetest.moc"
