#include "isegclientapi.h"
#include "version.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUFFER_SIZE 512
int  m_socketDescriptor = 0;
static pthread_mutex_t m_mutexSocket = PTHREAD_MUTEX_INITIALIZER;

/**
 * @brief Returns the library four digit version number as unsigned int value.
 *
 * This allows to check for a specific version number.
 * @since 1.1
 * @return The version number as unsigned int value, e.g. 0x01010000
 */
unsigned int iseg_getVersion(void)
{
	const unsigned int version[] = { VER_FILEVERSION };

	return (version[0] << 24) + (version[1] << 16) + (version[2] << 8) + version[3];
}

/**
 * @brief Returns the library four digit version number as null-terminated string.
 *
 * The digits are separated by dots.
 * @since 1.1
 * @return The version number, e.g. "1.1.0.3"
 */
const char *iseg_getVersionString(void)
{
	return VER_FILEVERSION_STR;
}

IsegResult iseg_connect(const char *name, const char *interface, void *reserved)
{
	UNREFERENCED_PARAMETER(reserved);

	if (m_socketDescriptor <= 0) {
		struct sockaddr_un address;

		address.sun_family = AF_LOCAL;
		strcpy(address.sun_path, UDS_FILE);
		if ((m_socketDescriptor = socket(PF_LOCAL, SOCK_STREAM, 0)) < 0) {
			//printf("Fehler bei socket()");// ...(%s)\n", strerror(errno));
			return ISEG_ERROR;
		}

		if (connect(m_socketDescriptor, (struct sockaddr *) &address, sizeof(address))) {
			//printf("Fehler bei connect");// bei connect() ...(%s)\n", strerror(errno));
			return ISEG_ERROR;
		}
	}

	char buf[BUFFER_SIZE];
	char recvBuf[3] = { 0 };
	int length = snprintf(buf, BUFFER_SIZE, "c%s,%s,",name, interface);
	pthread_mutex_lock(&m_mutexSocket);
	send(m_socketDescriptor, buf, length, 0);
	recv(m_socketDescriptor, recvBuf, 3, 0);
	pthread_mutex_unlock(&m_mutexSocket);
	return atoi(recvBuf);
}

IsegResult iseg_disconnect(const char *name)
{
	char buf[BUFFER_SIZE] = "d";
	char recvBuf[3];
	strncpy(buf + 1, name, INTERFACE_NAME_SIZE);
	pthread_mutex_lock(&m_mutexSocket);
	send(m_socketDescriptor, buf, strlen(buf), 0);
	recv(m_socketDescriptor, recvBuf, 3, 0);
	pthread_mutex_unlock(&m_mutexSocket);

	close(m_socketDescriptor);
	m_socketDescriptor = 0;
	return atoi(recvBuf);
}

/**
 * @brief writes an item value, specified by the item object, into the data cache
 * @param item - item struct
 * @return - 0 = ok
 *         - 1 = error
 */
IsegResult iseg_setItem(const char *name, const char *object, const char *value)
{
	char buf[BUFFER_SIZE];
	char recvBuf[3];
	int length = snprintf(buf, BUFFER_SIZE, "s%s,%s,%s",name, object, value);
	pthread_mutex_lock(&m_mutexSocket);
	send(m_socketDescriptor, buf, length, 0);
	recv(m_socketDescriptor, recvBuf, 3, 0);
	pthread_mutex_unlock(&m_mutexSocket);
	//if (strcmp(recvBuf, "0") != 0) {
	//	printf("Error client API %s\n", recvBuf);
	//}
	return atoi(recvBuf);
}

/**
 * @brief read an item value, specified by the item object, from the data cache
 * @param item object
 * @return isegItemValue { object, value, quality, time }
 */
IsegItem iseg_getItem(const char *name, const char *object)
{
	IsegItem item = EmptyIsegItem;
	char buf[BUFFER_SIZE];
	char recvBuf[BUFFER_SIZE];
	int length = snprintf(buf, BUFFER_SIZE, "g%s,%s",name,object);
	pthread_mutex_lock(&m_mutexSocket);
	send(m_socketDescriptor, buf, length, 0);
	recv(m_socketDescriptor, recvBuf, sizeof(IsegItem), 0);
	memcpy(&item, recvBuf, sizeof(IsegItem));
	pthread_mutex_unlock(&m_mutexSocket);
	return item;
}

/**
 * @brief read an item property, specified by the item object, from the data cache
 * @param item object
 * @return isegItemProperty { object, type, unit, access }
 */
IsegItemProperty iseg_getItemProperty(const char *name, const char *object)
{
	UNREFERENCED_PARAMETER(name);

	IsegItemProperty itemProperty = EmptyIsegItemProperty;
	char buf[BUFFER_SIZE] = "p";
	char recvBuf[BUFFER_SIZE];
	memcpy(buf+1, object, strlen(object));
	//printf("Client send Prop %s\n", buf);
	pthread_mutex_lock(&m_mutexSocket);
	send(m_socketDescriptor, buf, strlen(buf), 0);
	recv(m_socketDescriptor, recvBuf, sizeof(IsegItemProperty), 0);
	memcpy(&itemProperty, recvBuf, sizeof(IsegItemProperty));
	pthread_mutex_unlock(&m_mutexSocket);
	return itemProperty;
}
