#include "kcanapi.h"
#include "kdatacache.h"
#include "kloggerfifo.h"
#include "socket_can/mainthread.h"
#include "version.h"

#include <ctype.h>
#include <locale.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(SOCKET_CAN) && SOCKET_CAN==1
#	include <libsocketcan.h>
#endif

static char m_interface[128];
static int m_interfaceIndex = 0;

#define WRONG_INDEX 999
/**
 * @brief Returns the interfaxe index:
 * @return - index 0, 1, ..
 *         - error 999
 */
static int getInterfaceIndex()
{
	if (
			( (strncmp(m_interface, "can", 3) == 0) && (atoi(&m_interface[4]) < 16) ) ||
			strncmp(m_interface, "usb", 3) == 0 ||
			strncmp(m_interface, "com", 3) == 0
	) {
		return atoi(&m_interface[3]);
	}

	return WRONG_INDEX;
}

/**
 * @brief Returns the library four digit version number as unsigned int value.
 *
 * This allows to check for a specific version number.
 * @since 1.1
 * @return The version number as unsigned int value, e.g. 0x01010000
 */
unsigned int can_getVersion(void)
{
	const unsigned int version[] = { VER_FILEVERSION };

	return (version[0] << 24) + (version[1] << 16) + (version[2] << 8) + version[3];
}

static const char *can_getConnectedInterface(void)
{
	return m_interface;
}

/**
 * @brief Returns the library four digit version number as null-terminated string.
 *
 * The digits are separated by dots.
 * @since 1.1
 * @return The version number, e.g. "1.1.0.3"
 */
const char *can_getVersionString(void)
{
	return VER_FILEVERSION_STR;
}

/**
 * @brief Returns the state of the connected interface.
 * @param property
 * @return
 */
static char *can_getLineProperty(const char *property, uint8_t index)
{
	//TODO other interfaces
	if (
			strcmp(property, "Status") != 0 &&
			index == getInterfaceIndex()
	) {
		return STATE_UNDEFINED_STRING;
	}

#if defined(SOCKET_CAN) && SOCKET_CAN==1
	int state;

	/* first we check if we can restart the device at all */
	if ((can_get_state(can_getConnectedInterface(), &state)) < 0) {
		return CAN_STATE_SERIOUSLY_ERROR_STRING;
	} else {
		switch (state) {

		case CAN_STATE_ERROR_ACTIVE:
			return CAN_STATE_ERROR_ACTIVE_STRING;

		case CAN_STATE_ERROR_WARNING:
			return CAN_STATE_ERROR_WARNING_STRING;

		case CAN_STATE_ERROR_PASSIVE:
			return CAN_STATE_ERROR_PASSIVE_STRING;

		case CAN_STATE_BUS_OFF:
			return CAN_STATE_BUS_OFF_STRING;

		case CAN_STATE_STOPPED:
			return CAN_STATE_STOPPED_STRING;

		case CAN_STATE_SLEEPING:
			return CAN_STATE_SLEEPING_STRING;

		default:
			return STATE_UNDEFINED_STRING;
		}
	}

#else

	return 0;

#endif
}

uint32_t can_getBitRate(bool *ok)
{
#if defined(SOCKET_CAN) && (SOCKET_CAN==1)
	struct can_bittiming bt;

	if ((can_get_bittiming(can_getConnectedInterface(), &bt)) < 0) {
		*ok = false;
		return 0;
	} else {
		*ok = true;
		return bt.bitrate;
	}
#else
	UNREFERENCED_PARAMETER(ok);
#endif

	return 0;
}

/**
 * @brief Connects the specified interface to an iseg high voltage device.
 * @param interface - can0, can1, .. can15; usb0, usb1; com1; ip_address
 * @return - 0 = OK
 *         - 1 = ERROR
 */
IsegResult can_connect(const char *interface, bool firstConnection)
{
	strncpy(m_interface, interface, sizeof m_interface);
	//m_interfaceIndex = getInterfaceIndex();

	if (!firstConnection)
		return ISEG_OK;

	kdatacache_clear();

	setlocale(LC_ALL, "C");

	if (m_interfaceIndex == WRONG_INDEX)
		return ISEG_ERROR;

	if (!ksocketcan_start(m_interface))
		return ISEG_ERROR;

	kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanStatus);
	kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanBitRate);
	return ISEG_OK;
}

IsegResult can_disconnect(void)
{
	return ksocketcan_stop();
}

enum INDEX_NUMBER {
	LINE_INDEX = 0,
	MODULE_INDEX,
	CHANNEL_INDEX,
	ITEM_INDEX,
	BIT_ARRAY_INDEX,
	BYTE_ARRAY_INDEX,
	INDEX_SIZE
} INDEX_NUMBER;

static bool configureBitArrayIndex(uint16_t *objectAddress, char *p_object)
{
	char *colon = strchr(p_object, ':');
	if (!colon)
		return false;

	objectAddress[BIT_ARRAY_INDEX] = atoi(colon + 1);
	*colon = 0;
	return true;
}

static bool configureByteArrayIndex(uint16_t *objectAddress, char *p_object)
{
	char *openingBrace = strchr(p_object, '[');
	if (!openingBrace)
		return false;

	char *closingBrace = strchr(openingBrace, ']');
	if (closingBrace)
		*closingBrace = 0;
	objectAddress[BYTE_ARRAY_INDEX] = atoi(openingBrace + 1);
	*openingBrace = 0;
	return true;
}

enum {
	ObjectSystemParts = 1,
	ObjectLineParts,
	ObjectModuleParts,
	ObjectChannelParts,
	NoneObjectParts
};

uint8_t split(uint16_t *objectAddress, const char *itemObject, char separator)
{
	char object[FULLY_QUALIFIED_OBJECT_SIZE] = { 0 };
	char *p_object;
	uint8_t  separatorNumber = 0;
	uint8_t  pos = 0;
	uint8_t  idx = 0;
	uint16_t  address = 0;

	strncpy(object, itemObject, FULLY_QUALIFIED_OBJECT_SIZE);
	object[FULLY_QUALIFIED_OBJECT_SIZE - 1] = 0;
	p_object = object;
	objectAddress[BIT_ARRAY_INDEX] = NO_BIT_ARRAY_INDEX;
	for (; pos < FULLY_QUALIFIED_OBJECT_SIZE; pos++, idx++) {

		if (p_object[idx] == 0) {
			break;
		}

		if (separatorNumber >= ObjectChannelParts) {
			break;
		}
		if (!strchr(p_object, separator)) {
			if (++separatorNumber == ObjectChannelParts) {
				configureBitArrayIndex(objectAddress, p_object);
				objectAddress[ITEM_INDEX] = kdatacache_getChannelColumn(p_object);
				return separatorNumber;
			}
			if (separatorNumber == ObjectModuleParts) {
				if (!configureBitArrayIndex(objectAddress, p_object))
					configureByteArrayIndex(objectAddress, p_object);
				if (
						(objectAddress[MODULE_INDEX] >= CRATE_ADDRESS_OFFSET && objectAddress[MODULE_INDEX] < CRATE_ADDRESS_OFFSET + DataCacheMaxCrates) ||
						(objectAddress[MODULE_INDEX] >= CRATE_ECH238_ADDRESS_OFFSET && objectAddress[MODULE_INDEX] < CRATE_ECH238_ADDRESS_OFFSET + DataCacheMaxCrates)
				) {
					objectAddress[ITEM_INDEX] = kdatacache_getCrateColumn(p_object);
					return separatorNumber;
				}
				objectAddress[ITEM_INDEX] = kdatacache_getModuleColumn(p_object);
				return separatorNumber;
			}

			if (separatorNumber == ObjectLineParts) {
				configureBitArrayIndex(objectAddress, p_object);
				objectAddress[ITEM_INDEX] = kdatacache_getDeviceColumn(p_object);
				return separatorNumber;
			}

			if (separatorNumber == ObjectSystemParts) {
				if (!configureBitArrayIndex(objectAddress, p_object))
					configureByteArrayIndex(objectAddress, p_object);
				objectAddress[ITEM_INDEX] = kdatacache_getSystemColumn(p_object);
				return separatorNumber;
			}

			return separatorNumber;
		}

		if (p_object[idx] == separator) {
			if (idx == 0) {
				address = -1;		//alle angeschlossenen Datenpunkte
			}

			if (separatorNumber == 0 && address >= DataCacheMaxLines)
				return NoneObjectParts;
			if (
					separatorNumber == 1 &&
					(
						(address >= DataCacheMaxModules && address < CRATE_ADDRESS_OFFSET) ||
						(address >= (CRATE_ADDRESS_OFFSET + DataCacheMaxCrates) && address < CRATE_ECH238_ADDRESS_OFFSET) ||
						(address >= (CRATE_ECH238_ADDRESS_OFFSET + DataCacheMaxCrates))
					)
			)
				return NoneObjectParts;
			if (separatorNumber == 2 && address >= DataCacheMaxChannels)
				return NoneObjectParts;

			objectAddress[separatorNumber] = address;
			if (++separatorNumber >= 4) {
				break;
			}

			p_object = (p_object + idx + 1);

			address = 0;
			idx = -1;
			continue;
		}
		address *= 10;
		address += p_object[idx] - '0';
	}

	return NoneObjectParts;
}

static IsegResult setItemValue(const IsegItem *item, uint16_t objectAddress[5], uint16_t number)
{
	uint16_t line = objectAddress[LINE_INDEX];
	uint16_t device = objectAddress[MODULE_INDEX];
	uint16_t channel = objectAddress[CHANNEL_INDEX];
	uint16_t column = objectAddress[ITEM_INDEX];
	uint8_t bit = objectAddress[BIT_ARRAY_INDEX];

	if (number == ObjectChannelParts) {
		DataType type = kdatacache_getChannelDataTypeEnum(column);
		if (type == DataTypeR4) {
			return kdatacache_setChannelValueR4(atof(item->value), line, device, channel, column);
		}

		return kdatacache_setChannelValueUI4(atoi(item->value), line, device, channel, column, bit);
	}

	if (number == ObjectModuleParts) {
		if (device >= CRATE_ADDRESS_OFFSET) {
			bool isEch238 = false;
			CrateCache *crateCacheValue;

			if (device >= CRATE_ECH238_ADDRESS_OFFSET) {
				isEch238 = kdatacache_isEch238(line, device - CRATE_ECH238_ADDRESS_OFFSET);
				if (!isEch238) {
					return ISEG_ERROR;
				}
				device -= CRATE_ADDRESS_OFFSET;	// ECH238
			}
			device -= CRATE_ADDRESS_OFFSET;

			if (!kdatacache_isCrateAlive(line, device, isEch238))
				return ISEG_ERROR;

			crateCacheValue = kdatacache_crateCacheValue(line, device, isEch238);
			DataType type = kdatacache_getCrateDataTypeEnum(column);
			if (type == DataTypeUI4 || type == DataTypeUI1 || type == DataTypeBool) {
				if (kdatacache_setCrateValueUI4(atoi(item->value), crateCacheValue, column, bit) == ISEG_OK) {
					kdataCache_setCrateCacheStateNewValue(line, device, column, isEch238);
					return ISEG_OK;
				}
			}

			return ISEG_ERROR;

		}

		if (!kdatacache_isModulAlive(line, device))
			return ISEG_ERROR;

		DataType type = kdatacache_getModuleDataTypeEnum(column);
		if (type == DataTypeR4) {
			return kdatacache_setModuleValueR4(atof(item->value), line, device, column);
		}

		if (type == DataTypeString) {
			return kdatacache_setModuleValueString(item->value, line, device, column);
		}

		return kdatacache_setModuleValueUI4(atoi(item->value), line, device, column, bit);
	}

	if (number == ObjectSystemParts) {
		DataType type = kdatacache_getSystemDataTypeEnum(column);
		if (type == DataTypeUI4) {
			return kdatacache_setSystemValue(atoi(item->value), column);
		}

		if (type == DataTypeString) {
			return kdatacache_setSystemValueString(item->value, column);
		}
	}

	return ISEG_ERROR;
}

/**
 * @brief writes an item value, specified by the item object, into the data cache
 * @param item - item struct
 * @return - 0 = ok
 *         - 1 = error
 */
IsegResult can_setItemValue(const IsegItem *item)
{
	IsegResult result;
	uint16_t  objectAddress[INDEX_SIZE] = { 0 };

	int number = split(objectAddress, item->object, '.');

	if (objectAddress[LINE_INDEX] < m_interfaceIndex) {
		//printf("Error API %s %s\n", item->object, item->value);
		return ISEG_ERROR;
	}

	objectAddress[LINE_INDEX] -= m_interfaceIndex;

	if (number == NoneObjectParts) {
		//printf("Error API %s %s\n", item->object, item->value);
		return ISEG_ERROR;
	}

	kloggerfifo_setApiInfo(LogApiWrite, "APITX", item);
	kmutex_lock();
	result = setItemValue(item, objectAddress, number);
	kmutex_unlock();

	//if (result != ISEG_OK) {
	//	printf("Error API %s %s\n", item->object, item->value);
	//}
	return result;
}

static void itemValue(IsegItem *item, uint16_t *objectAddress, uint16_t number)
{
	uint16_t line = objectAddress[LINE_INDEX];
	uint16_t device = objectAddress[MODULE_INDEX];
	uint16_t channel = objectAddress[CHANNEL_INDEX];
	uint16_t column = objectAddress[ITEM_INDEX];
	uint8_t bit = objectAddress[BIT_ARRAY_INDEX];
	uint8_t byte = objectAddress[BYTE_ARRAY_INDEX];

	if (number == ObjectChannelParts) {
		if (objectAddress[ITEM_INDEX] >= DataCacheColumnChannelSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		DataType type = kdatacache_getChannelDataTypeEnum(column);
		strncpy(item->timeStampLastRefreshed,
				kdatacache_getChannelTimeStampLastRefreshed(line, device, channel, column), TIME_SIZE);
		strncpy(item->timeStampLastChanged,
				kdatacache_getChannelTimeStampLastChanged(line, device, channel, column), TIME_SIZE);
		strncpy(item->quality,
				kdatacache_getChannelQuality(line, device, channel, column), QUALITY_SIZE);

		if (type == DataTypeR4) {
			float value = kdatacache_getChannelValueR4(line, device, channel, column);
			snprintf(item->value, VALUE_SIZE, "%g", value);
			return;
		}

		uint32_t value = kdatacache_getChannelValueUI4(line, device, channel, column, bit);
		snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)value);
		return;
	}

	if (number == ObjectModuleParts) {

		if (byte >= INTERNAL_TEMPERATURE_ARRAY_SIZE) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		if (device >= CRATE_ADDRESS_OFFSET) {
			bool requestEch238 = false;
			CrateCache *crateCacheValue;

			if (column >= DataCacheColumnCrate4xaSize) {
				strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
				return;
			}

			if (device >= CRATE_ECH238_ADDRESS_OFFSET) {
				requestEch238 = true;
				device -= CRATE_ADDRESS_OFFSET;	// ECH238
			}
			device-= CRATE_ADDRESS_OFFSET;

			strncpy(item->quality, kdatacache_getCrateQuality(line, device, column, requestEch238), QUALITY_SIZE);
			crateCacheValue = kdatacache_crateCacheValue(line, device, requestEch238);

			DataType type = kdatacache_getCrateDataTypeEnum(column);
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getCrateTimeStampLastRefreshed(line, device, column, requestEch238), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getCrateTimeStampLastChanged(line, device, column, requestEch238), TIME_SIZE);

			if (type == DataTypeR4) {
				float value = kdatacache_getCrateValueR4(crateCacheValue, column, byte);
				snprintf(item->value, VALUE_SIZE, "%g", value);
				return;
			}

			if (type == DataTypeUI4 || type == DataTypeUI1 || type == DataTypeBool) {
				uint32_t value = kdatacache_getCrateValueUI4(crateCacheValue, column, bit);
				snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)value);
				return;
			}

			if (type == DataTypeString) {
				strcpy(item->value, kdatacache_getCrateValueString(crateCacheValue, column));
				return;
			}

			strcpy(item->quality, ISEG_ITEM_QUALITY_ERROR);
			return;
		}

		if (column >= DataCacheColumnModuleSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		DataType type = kdatacache_getModuleDataTypeEnum(column);
		strncpy(item->timeStampLastRefreshed,
				kdatacache_getModuleTimeStampLastRefreshed(line, device, column), TIME_SIZE);
		strncpy(item->timeStampLastChanged,
				kdatacache_getModuleTimeStampLastChanged(line, device, column), TIME_SIZE);
		strncpy(item->quality, kdatacache_getModuleQuality(line, device, column), QUALITY_SIZE);

		if (type == DataTypeR4) {
			float value = kdatacache_getModuleValueR4(line, device, column, byte);
			snprintf(item->value, VALUE_SIZE, "%g", value);
			return;
		}

		if (type == DataTypeUI4 || type == DataTypeBool) {
			uint32_t value = kdatacache_getModuleValueUI4(line, device, column, bit);
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)value);
			return;
		}

		if (type == DataTypeString) {
			strcpy(item->value, kdatacache_getModuleValueString( line, device, column));
			return;
		}
	}

	if (number == ObjectLineParts) {
		if (column >= DataCacheColumnLineSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		if (column == DataCacheColumnLineCanStatus) {
			strcpy(item->value, can_getLineProperty(kdatacache_getLinePropertyString(column), line));
			kdatacache_createLineTimeStamp(line, column);
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getLineTimeStampLastRefreshed(line, column), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(DataCacheColumnSystemCanStatus), TIME_SIZE);
			strncpy(item->quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			return;
		}
		if (column == DataCacheColumnLineCanBitRate) {
			bool ok = false;
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)can_getBitRate(&ok));
			kdatacache_createLineTimeStamp(line, column);
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getLineTimeStampLastRefreshed(line, column), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(DataCacheColumnSystemCanBitRate), TIME_SIZE);
			if (ok) {
				strncpy(item->quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			} else {
				strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			}
			return;
		}

		DataType type = kdatacache_getLineDataTypeEnum(column);
		strncpy(item->timeStampLastRefreshed,
				kdatacache_getLineTimeStampLastRefreshed(line, column), TIME_SIZE);
		strncpy(item->timeStampLastChanged,
				kdatacache_getLineTimeStampLastChanged(line, column), TIME_SIZE);
		strncpy(item->quality,
				kdatacache_getLineQuality(line, column), QUALITY_SIZE);

		if (type == DataTypeUI4) {
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)kdatacache_getDeviceValue(line, column));
			return;
		}

		if (type == DataTypeString) {
			strcpy(item->value, kdatacache_getLineValueString(line, column, bit));
			return;
		}
	}

	if (number == ObjectSystemParts) {

		if (column >= DataCacheColumnSystemSize) {
			strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			return;
		}

		kdatacache_createSystemTimeStamp(false, column);
		if (column == DataCacheColumnSystemCanStatus) {
			strcpy(item->value, can_getLineProperty(kdatacache_getSystemPropertyString(column), 0));	//TODO
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getSystemTimeStampLastRefreshed(column), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(column), TIME_SIZE);

			strncpy(item->quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			return;
		}
		if (column == DataCacheColumnSystemCanBitRate) {
			bool ok = false;
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)can_getBitRate(&ok));
			strncpy(item->timeStampLastRefreshed,
					kdatacache_getSystemTimeStampLastRefreshed(column), TIME_SIZE);
			strncpy(item->timeStampLastChanged,
					kdatacache_getSystemTimeStampLastChanged(column), TIME_SIZE);
			if (ok) {
				strncpy(item->quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			} else {
				strncpy(item->quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
			}
			return;
		}

		DataType type = kdatacache_getSystemDataTypeEnum(column);
		strncpy(item->timeStampLastChanged, kdatacache_getSystemTimeStampLastChanged(column), TIME_SIZE);
		strncpy(item->timeStampLastRefreshed, kdatacache_getSystemTimeStampLastRefreshed(column), TIME_SIZE);
		strncpy(item->quality, kdatacache_getSystemQuality(column), QUALITY_SIZE);

		if (type == DataTypeUI4) {
			snprintf(item->value, VALUE_SIZE, "%lu", (unsigned long)kdatacache_getSystemValue(column));
			return;
		}

		if (type == DataTypeString) {
			strcpy(item->value, kdatacache_getSystemValueString(column, bit));
			return;
		}

	}

	strcpy(item->quality, ISEG_ITEM_QUALITY_ERROR);
	return;
}

/**
 * @brief read an item value, specified by the item object, from the data cache
 * @param item object
 * @return isegItemValue { object, value, quality, time }
 */
IsegItem can_getItemValue(const char *object)
{
	IsegItem item = EmptyIsegItem;
	uint16_t objectAddress[INDEX_SIZE] = { 0 };

	strncpy(item.object, object, FULLY_QUALIFIED_OBJECT_SIZE);
	int number = split(objectAddress, object, '.');

	strncpy(item.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
	if (objectAddress[LINE_INDEX] < m_interfaceIndex) {
		return item;
	}

	objectAddress[LINE_INDEX] -= m_interfaceIndex;

	if (number == NoneObjectParts) {
		strcpy(item.quality, ISEG_ITEM_QUALITY_ERROR);
		return item;
	}

	kmutex_lock();
	itemValue(&item, objectAddress, number);
	kmutex_unlock();

	kloggerfifo_setApiInfo(LogApiWrite, "APIRX", &item);

	return item;
}

/**
 * @brief read an item property, specified by the item object, from the data cache
 * @param item object
 * @return isegItemProperty { object, type, unit, access }
 */
IsegItemProperty can_getItemProperty(const char *object)
{
	IsegItemProperty itemProperty = EmptyIsegItemProperty;
	uint16_t         objectAddress[INDEX_SIZE] = { 0 };

	strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);

	strncpy(itemProperty.object, object, FULLY_QUALIFIED_OBJECT_SIZE);
	int number = split(objectAddress, object, '.');
	uint16_t line = objectAddress[LINE_INDEX];
	uint16_t device = objectAddress[MODULE_INDEX];
	uint16_t column = objectAddress[ITEM_INDEX];
	uint8_t bit = objectAddress[BIT_ARRAY_INDEX];

	if (objectAddress[LINE_INDEX] < m_interfaceIndex)
		return itemProperty;

	objectAddress[LINE_INDEX] -= m_interfaceIndex;

	if (number < ObjectSystemParts || number > ObjectChannelParts) {
		return itemProperty;
	}

	if (number == ObjectChannelParts && column < DataCacheColumnChannelSize) {
		strncpy(itemProperty.type, kdatacache_getChannelDataTypeString(column, bit), DATA_TYPE_SIZE);
		strncpy(itemProperty.unit, kdatacache_getChannelUnitString(column), UNIT_SIZE);
		strncpy(itemProperty.access, kdatacache_getChannelAccessString(column), ACCESS_SIZE);
		strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		return itemProperty;
	}

	if (number == ObjectModuleParts) {
		if (device >= CRATE_ADDRESS_OFFSET) {
			bool isEch238 = false;
			if (device >= CRATE_ECH238_ADDRESS_OFFSET) {
				isEch238 = kdatacache_isEch238(line, device - CRATE_ECH238_ADDRESS_OFFSET);
				if (!isEch238) {
					return itemProperty;
				}
				device -= CRATE_ADDRESS_OFFSET;	// ECH238
			}
			device -= CRATE_ADDRESS_OFFSET;

			if (!kdatacache_isCrateAlive(line, device, isEch238))
				return itemProperty;

			if (column < DataCacheColumnCrate4xaSize) {
				strncpy(itemProperty.type, kdatacache_getCrateDataTypeString(column, bit),
						DATA_TYPE_SIZE);
				strncpy(itemProperty.unit, kdatacache_getCrateUnitString(column), UNIT_SIZE);
				strncpy(itemProperty.access, kdatacache_getCrateAccessString(column), ACCESS_SIZE);
				strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			}

			return itemProperty;
		}

		if (column < DataCacheColumnModuleSize) {
			strncpy(itemProperty.type, kdatacache_getModuleDataTypeString(column, bit), DATA_TYPE_SIZE);
			strncpy(itemProperty.unit, kdatacache_getModuleUnitString(column), UNIT_SIZE);
			strncpy(itemProperty.access, kdatacache_getModuleAccessString(column), ACCESS_SIZE);
			strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			return itemProperty;
		}
	}

	if (number == ObjectLineParts) {
		strncpy(itemProperty.type, kdatacache_getLineDataTypeString(column), DATA_TYPE_SIZE);
		strncpy(itemProperty.unit, kdatacache_getLineUnitString(column), UNIT_SIZE);
		strncpy(itemProperty.access, kdatacache_getLineAccessString(column), ACCESS_SIZE);
		strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		return itemProperty;
	}

	if (number == ObjectSystemParts) {
		strncpy(itemProperty.type, kdatacache_getSystemDataTypeString(column), DATA_TYPE_SIZE);
		strncpy(itemProperty.unit, kdatacache_getSystemUnitString(column), UNIT_SIZE);
		strncpy(itemProperty.access, kdatacache_getSystemAccessString(column), ACCESS_SIZE);
		strncpy(itemProperty.quality, ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		return itemProperty;
	}

	return itemProperty;
}
