#ifndef KCAN_H
#define KCAN_H

#include <stdint.h>
#include <stdbool.h>

typedef union KCanId {
	uint16_t w;
	struct KCanIdBits {
		uint16_t readRequest: 1;   ///< Indicates an EDCP or DCP read request
		uint16_t dcpExtended: 1;   ///< Indicates an "Extended DCP" command frame
		uint16_t antiCollision: 1; ///< Set this bit to one when transmitting a frame from the module
		uint16_t address: 6;       ///< The modules node address (0...63)
		uint16_t passive: 1;       ///< This bit is always one except for "Active Error Messages"
		uint16_t crate: 1;         ///< Every crate has this bit set to one
		uint16_t configuration: 1; ///< Extended ID: ECH4XA crate configuration sequence
		uint16_t addressOffset: 4; ///< Extended ID: ECH4XA crate address offset (CAN line)
	} bit;
} KCanId;

typedef struct KCanMessage {
	KCanId id;           ///< The 11bit CAN identifier
	uint16_t dataId;     ///< The iseg Data ID (EDCP: 16bit, DCP: 8bit)
	uint8_t data[7];     ///< The iseg EDCP or DCP data field
	struct KCanMessageBits {
		uint8_t dlc:  4;        ///< The CAN Data Length Code (0...8}
		uint8_t busy: 1;        ///< Indicates the message buffer is occupied
		uint8_t rtr:  1;        ///< Indicates a CAN Remote Transmission Request frame
		uint8_t edcp: 1;        ///< Indicates a iseg EDCP frame

		uint8_t extendedId:  1; ///< Indicates a 29 bit (extended) identifier
	} bit;
} KCanMessage;

#ifndef EmptyCanMessage
#define EmptyCanMessage { { 0 }, 0, { 0 }, { 0 } }
#endif

typedef enum KCanBaudRate {
	CanBaudRateUnknown = 0,
	CanBaudRate20kbps  = 20,
	CanBaudRate50kbps  = 50,
	CanBaudRate100kbps = 100,
	CanBaudRate125kbps = 125,
	CanBaudRate250kbps = 250,
	CanBaudRate500kbps = 500,
	CanBaudRate1000kbps = 1000
} KCanBaudRate;

typedef enum KCanStandardIdentifierBits {
	CanId0               = 0x001,
	CanId0_readRequest   = CanId0,
	CanId1               = 0x002,
	CanId1_dcpExtended   = CanId1,
	CanId2               = 0x004,
	CanId2_antiCollision = CanId2,
	CanId3               = 0x008,
	CanId3_address0      = CanId3,
	CanId4               = 0x010,
	CanId4_address1      = CanId4,
	CanId5               = 0x020,
	CanId5_address2      = CanId5,
	CanId6               = 0x040,
	CanId6_address3      = CanId6,
	CanId7               = 0x080,
	CanId7_address4      = CanId7,
	CanId8               = 0x100,
	CanId8_address5      = CanId8,
	CanId9               = 0x200,
	CanId9_passive       = CanId9,
	CanId10              = 0x400,
	CanId10_crate        = CanId10,
	CanIdAddressMask     = 0x1F1
} KCanStandardIdentifierBits;

bool isGroupDataId(uint16_t dataid);
void kcan_noMasterCrate(void);
bool kcan_logged(uint16_t line, uint16_t address);
KCanMessage kcan_setLogOn(uint16_t line, uint16_t address, bool boardB);
KCanMessage kcan_createStandardEdcpMessage(KCanId id, bool boardB, bool extendedId, int dataId, int dlc, const uint8_t * const data);
KCanMessage kcan_createStandardDcpMessage(KCanId id, bool boardB, bool extendedId, int dataId, int dlc, const uint8_t * const data);
bool kcan_isChannelDataId(uint16_t dataid);
KCanMessage kcan_setLogOff(uint16_t line, uint16_t address, bool boardB);
bool kcan_cacheAllAnwersReceived(uint16_t line, uint16_t address);
void kcrate_setAlive(bool alive, uint16_t line, uint16_t crate);
bool kcan_cacheAlive(uint16_t line, uint16_t address);
bool kcan_cacheSetValuesUpdated(uint16_t line, uint16_t address);
KCanMessage kcan_createAntiCollisionMessage(uint16_t line, uint16_t module, bool boardB);
uint8_t kcan_moduleAddressToSlotNumber(uint16_t line, uint16_t address, bool *boardB);
uint16_t kcan_slotNumberToDoubleModuleAddress(uint16_t slot, bool boardB);
uint16_t kcan_slotNumberToSingleModuleAddress(uint16_t slot);
bool kcan_isDoubleModule(uint16_t line, uint16_t module);

#endif // KCAN_H
