#include "kcan.h"
#include "isegcommon.h"
#include "kloggerfifo.h"

#include <pthread.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static uint32_t m_level;
static pthread_mutex_t m_mutex = PTHREAD_MUTEX_INITIALIZER;

extern KLoggerFifo m_fifo;

/**
 * @brief Initializes a Logger FIFO object.
 *
 * The FIFO object and the given buffer are initialized with zero.
 * @param self Pointer to a FIFO object.
 * @param buffer Pointer to an char array logger information
 * @param capacity The number of elements in the KLoggerMessage array
 */
void kloggerfifo_init(KLoggerFifo *self, KLoggerMessage *buffer, uint8_t capacity)
{
	memset(self, 0, sizeof(KLoggerFifo));
	self->capacity = capacity;
	self->data = buffer;
	memset(self->data, 0, sizeof(KLoggerMessage) * capacity);
}

/**
 * @brief Returns the maximum number of logger information that can be queued.
 * @param self Pointer to a FIFO object.
 * @return The FIFOs capacity in elements.
 */
static uint8_t kloggerfifo_capacity(KLoggerFifo *self)
{
	return self->capacity;
}

/**
 * @brief Returns the number of logger information currently in the FIFO.
 * @param self Pointer to a FIFO object.
 * @return The number of elements currently used.
 */
static uint8_t kloggerfifo_size(KLoggerFifo *self)
{
	return self->size;
}

/**
 * @brief Returns true if the FIFO is empty.
 * @param self Pointer to a FIFO object.
 * @return True if the FIFO is empty, false otherwise
 */
static bool kloggerfifo_isEmpty(KLoggerFifo *self)
{
	return kloggerfifo_size(self) == 0;
}

/**
 * @brief Returns true if the FIFO is half full.
 * @param self Pointer to a FIFO object.
 * @return True if the FIFO is half full, false otherwise
 */
static bool kloggerfifo_isHalfFull(KLoggerFifo *self)
{
	return kloggerfifo_size(self) >= kloggerfifo_capacity(self) / 2;
}

/**
 * @brief Returns true if the FIFO is full.
 * @param self Pointer to a FIFO object.
 * @return True if the FIFO is full, false otherwise
 */
static bool kloggerfifo_isFull(KLoggerFifo *self)
{
	return kloggerfifo_size(self) == kloggerfifo_capacity(self) - 1;
}

/**
 * @brief Returns the next logger message from the FIFO.
 *
 * The next message is the one that was enqueued first.
 * @see kloggerfifo_enqueue
 * @param self Pointer to a FIFO object.
 * @return Pointer to the next logger message or NULL if the FIFO is empty.
 */
static KLoggerMessage * kloggerfifo_dequeue(KLoggerFifo *self)
{
	if (kloggerfifo_isEmpty(self))
		return NULL;

	--self->size;
	if (++self->head == kloggerfifo_capacity(self))
		self->head = 0;

	return &self->data[self->head];
}

/**
 * @brief Enqueues a logger message to the FIFO.
 *
 * @see kloggerfifo_dequeue
 * @param self Pointer to a FIFO object.
 * @param message Pointer to the message to enqueue
 * @return True if the message was successfully enqueued,
 * false if the buffer is full.
 */
static bool kloggerfifo_enqueue(KLoggerFifo *self, KLoggerMessage *message)
{
	if (kloggerfifo_isFull(self))
		return false;

	++self->size;

	pthread_mutex_lock(&m_mutex);
	if (++self->tail == kloggerfifo_capacity(self))
		self->tail = 0;
	uint8_t tail = self->tail;
	pthread_mutex_unlock(&m_mutex);

	memcpy(&self->data[tail], message, sizeof(KLoggerMessage));
	return true;
}

/**
 * @brief Returns true if the given logger message is already in the FIFO.
 * @param self Pointer to a FIFO object.
 * @param message Pointer to the message that should be checked
 * @return True if the message is already in the FIFO, false otherwise.
 */
static bool kloggerfifo_contains(KLoggerFifo *self, KLoggerMessage *info)
{
	uint8_t count = kloggerfifo_capacity(self);
	uint8_t i;
	for (i = 0; i < count; ++i) {
		if (memcmp(&self->data[i], info, sizeof(KLoggerMessage)) == 0)
			return true;
	}

	return false;
}

void kloggerfifo_setLevel(uint32_t level)
{
	m_level = level;
}

void kloggerfifo_setMessage(uint32_t event, char *info, uint16_t line, uint16_t device)
{
	if ((m_level & event) == 0)
		return;

	KLoggerMessage message;
	snprintf(message, LOGGER_MESSAGE_SIZE, "%02x %s %hu.%hu!", event, info, line, device);
	kloggerfifo_enqueue(&m_fifo, &message);
}

static const char qualityString[5][15] = {
	"QUA INVALID",
	"QUA INIT",
	"QUA OK",
	"QUA BAD",
	"QUA ERROR"
};

void kloggerfifo_setApiInfo(uint32_t event, char *info, const IsegItem *item)
{
	if ((m_level & event) == 0)
		return;

	KLoggerMessage message;
	snprintf(message, LOGGER_MESSAGE_SIZE, "%02x %s [ %s = %s ] %s",
			 event, info, item->object, item->value, qualityString[atoi(item->quality)]);
	kloggerfifo_enqueue(&m_fifo, &message);
}

void kloggerfifo_setFrame(uint32_t event, char *info, const KCanMessage *frame, int count, uint32_t answers)
{
	if ((m_level & event) == 0)
		return;

	int i;
	char msg[LOGGER_MESSAGE_SIZE];
	int size = frame->bit.dlc;
	int n = snprintf(msg, LOGGER_MESSAGE_SIZE, "%02x %s %04X [", m_level & event, info, frame->id.w);
	if (frame->bit.edcp) {
		size -= 2;
		n += snprintf(msg + n, LOGGER_MESSAGE_SIZE - n, " %04X", frame->dataId);
	} else {
		size -= 1;
		n += snprintf(msg + n, LOGGER_MESSAGE_SIZE - n, " %02X", frame->dataId);
	}
	for (i = 0; i < size; ++i)
		n += snprintf(msg + n,LOGGER_MESSAGE_SIZE - n, " %02X", frame->data[i]);
	n += snprintf(msg + n, LOGGER_MESSAGE_SIZE - n, "] ");
	n += snprintf(msg + n, LOGGER_MESSAGE_SIZE - n, "number = %lu expected answers = %lu", (unsigned long)count, (unsigned long)answers);
	msg[n] = 0;
	kloggerfifo_enqueue(&m_fifo, &msg);
}

bool kloggerfifo_getInfo(char **message)
{
	if (kloggerfifo_isEmpty(&m_fifo))
		return false;

	*message = (char*)kloggerfifo_dequeue(&m_fifo);

	return message != 0;
}
