#include "isegapi.h"
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>

#ifdef SYSTEMD_SOCKET_ACTIVATION
#include "systemd/sd-daemon.h"
#endif

#define MAX_LEN  4096

enum { Debug = 0 };

pthread_t tid;

static int pidfd, sockfd;
static const char pidfile[120] = "/var/run/iseghal/iseghalserver.pid";

static void shutdownDaemon(int exitCode)
{
	// TODO: Cleanly disconnect all active clients before closing the socket

	printf("Terminating server\n");
	close(sockfd);
	close(pidfd);
	if (unlink(pidfile)) {
		fprintf(stderr, "Could not delete PID lock file '%s' (%s).",
			   pidfile, strerror(errno));
	}
	closelog();
	exit(exitCode);
}

static void signalhandler(int signal)
{
	switch (signal) {
	case SIGHUP:
		syslog(LOG_WARNING, "Received SIGHUP signal.\n");
		break;
	case SIGPIPE:
		syslog(LOG_ERR, "Received SIGPIPE.\n");
		shutdownDaemon(EXIT_FAILURE);
		break;
	case SIGINT:
	case SIGKILL:
	case SIGTERM:
		shutdownDaemon(EXIT_SUCCESS);
		break;
	default:
		syslog(LOG_WARNING, "Unhandled signal %s\n", strsignal(signal));
		break;
	}
}

static bool isolateStringFromMessage(char *msg, u_int8_t number, char *p, size_t n)
{
	if (!p)
		return false;

	char * m = msg;
	if (number == 2) {
		m = strchr(msg, ',') + 1;
	} else if (number == 3) {
		m = strrchr(msg, ',') + 1;
	}

	strncpy(p, m, n);
	char *colonPos = strchr(p, ',');
	if (colonPos)
		*colonPos = 0;

	return true;
}

static void handleSocketInitialization()
{
	struct sockaddr_un serverAddress;
#ifndef SYSTEMD_SOCKET_ACTIVATION
	//initialize socket
	if ((sockfd = socket (AF_LOCAL, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr, "Error opening socket...(%s)\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	serverAddress.sun_family = AF_LOCAL;
	strcpy(serverAddress.sun_path, UDS_FILE);

	unlink(UDS_FILE);

	if (bind(sockfd, (struct sockaddr *) &serverAddress, sizeof (serverAddress))) {
		fprintf(stderr, "Error binding socket ...(%s)\n", strerror(errno));
		exit(EXIT_FAILURE);
	}
	printf("Successfully created socket\n");

	if (listen(sockfd, 5)) {
		fprintf(stderr, "Error listen() ...(%s)\n", strerror(errno));
		shutdownDaemon(EXIT_FAILURE);
	}
#else
	int n;
	printf("Listen for systemd passing me FDs...");
	n = sd_listen_fds(0); //check for file descriptors passed by the init system
	if (n < 0) {
		fprintf(stderr, "Failed to acquire sockets from systemd... %s", strerror(-n));
		shutdownDaemon(EXIT_FAILURE);
	}
	if (n == 0) {
		fprintf(stderr, "No file descriptors received.\n");
		shutdownDaemon(EXIT_FAILURE);
	}
	if (n > 1) {
		fprintf(stderr, "Too many file descriptors received.\n");
		shutdownDaemon(EXIT_FAILURE);
	}
	sockfd = SD_LISTEN_FDS_START;
	printf("Descriptor successfully passed.");
#endif
}

void *serviceClient(void* arg)
{
	pthread_detach(pthread_self());

	int fd = *(int *)arg;
	char m_sessionName[INTERFACE_NAME_SIZE] = { 0 };

	while(1)
	{ //socket offen lassen, bis nichts mehr kommt
		ssize_t n;
		bool res = true;
		char isegHalResult[3];
		char msg[MAX_LEN];
		char name[INTERFACE_NAME_SIZE];
		char interface[INTERFACE_TYPE_SIZE];
		char object[OBJECT_NAME_SIZE];
		char value[VALUE_SIZE];
		IsegItem tmpItem = EmptyIsegItem;
		IsegItemProperty tmpProp = EmptyIsegItemProperty;

		n = recv(fd, msg, MAX_LEN, 0);
		if (n <= 0) {
			break; //raus, socket zu machen
		}

		msg[n] = 0;
		switch (msg[0]) {
		case 'c': {	//iseg_connect
			res &= isolateStringFromMessage(msg + 1, 1, name, INTERFACE_NAME_SIZE);
			res &= isolateStringFromMessage(msg + 1, 2, interface, INTERFACE_TYPE_SIZE);

			if (Debug) {
				syslog(LOG_INFO, "iseg_connect(%s,%s,NULL", name, interface);
			}
			IsegResult connectResult = (res) ? iseg_connect(name, interface, NULL) : ISEG_ERROR;
			if (connectResult != ISEG_OK) {
				if (connectResult == ISEG_ERROR) {
					syslog(LOG_WARNING, "nonzero return value on iseg_connect. Return: ISEG_ERROR %s", interface);
				} else if (connectResult == ISEG_WRONG_SESSION_NAME) {
					syslog(LOG_WARNING, "nonzero return value on iseg_connect. Return: WRONG_SESSION_NAME %s", name);
				} else {
					syslog(LOG_WARNING, "nonzero return value on iseg_connect. Return: %d\n", connectResult);
				}
			} else if (m_sessionName[0] == 0) {
				strncpy(m_sessionName, name, INTERFACE_NAME_SIZE);
			}

			snprintf(isegHalResult, sizeof(isegHalResult), "%d", connectResult);
			send(fd, isegHalResult, 2, 0);
			break;
		}
		case 'd': {	//iseg_disconnect
			res &= isolateStringFromMessage(msg + 1, 1, name, INTERFACE_NAME_SIZE);
			IsegResult disconnectResult = (res) ? iseg_disconnect(name) : ISEG_ERROR;
			if (disconnectResult != ISEG_OK) {
				syslog(LOG_WARNING, "nonzero return value on iseg_disconnect. Return: %d\n", disconnectResult);
			}
			snprintf(isegHalResult, sizeof(isegHalResult), "%d", disconnectResult);
			send(fd, isegHalResult, 2, 0);
			break;
		}
		case 'g':	//iseg_getItem
			isolateStringFromMessage(msg + 1, 1, name, INTERFACE_NAME_SIZE);
			isolateStringFromMessage(msg + 1, 2, object, OBJECT_NAME_SIZE);
			tmpItem = iseg_getItem(name, object);
			send(fd, &tmpItem, sizeof(IsegItem), 0);
			break;
		case 's': {	//iseg_setItem
			if (n > OBJECT_NAME_SIZE + VALUE_SIZE) {
				syslog(LOG_WARNING, "wrong message length %ld %s\n", n, msg);
				if ((size_t)n > (sizeof(IsegItem) + 1))
					syslog(LOG_WARNING, "Second message %s\n", &msg[sizeof(IsegItem) + 1]);
				continue;
			}

			res &= isolateStringFromMessage(msg + 1, 1, name, INTERFACE_NAME_SIZE);
			res &= isolateStringFromMessage(msg + 1, 2, object, OBJECT_NAME_SIZE);
			res &= isolateStringFromMessage(msg + 1, 3, value, INTERFACE_TYPE_SIZE);
			IsegResult setResult = (res) ? iseg_setItem(name, object, value) : ISEG_ERROR;
			if (setResult != ISEG_OK) {
				syslog(LOG_WARNING, "nonzero return value on iseg_setItem. Return: %d\n", setResult);
			}
			snprintf(isegHalResult, sizeof(isegHalResult), "%d", setResult);
			send(fd, isegHalResult, strlen(isegHalResult), 0);
			break;
		}
		case 'p':	//iseg_getItemProperty
			//printf("empfangen Prop: %s\n", msg+1);
			tmpProp = iseg_getItemProperty(name, msg+1);
			send(fd, &tmpProp, sizeof(IsegItemProperty), 0);
			//printf("gesendet Prop: %s\n", tmpProp.object);
			break;
		default:
			send(fd, "", 2, 0);
			syslog(LOG_WARNING, "unknown message %s\n", msg);
		}
	}

	iseg_disconnect(m_sessionName);
	syslog(LOG_INFO, "Client disconnected.\n");
	close(fd);
	pthread_exit(NULL);
	return NULL;
}

int main (int argc, char *argv[])
{
	signal(SIGINT,  &signalhandler);
	signal(SIGHUP,  &signalhandler);
	signal(SIGTERM, &signalhandler);
	signal(SIGKILL, &signalhandler);
	signal(SIGPIPE, &signalhandler);

	int daemonize = 1;
	if (argc == 2 && strncmp(argv[1], "--fg", 4) == 0) {
		daemonize = 0;
		printf("Running isegHalServer as foreground process.\n");
	}

	handleSocketInitialization();
	if (daemonize) {
		/* Our process ID */
		pid_t pid;
		/* Fork off the parent process */
		pid = fork();
		if (pid < 0) {
			exit(EXIT_FAILURE);
		}
		/* If we got a good PID, then we can exit the parent process. */
		if (pid > 0) {
			exit(EXIT_SUCCESS);
		}

		/* Change the file mode mask */
		umask(0);
	}

	/* Open any logs here */
	openlog("isegHalServer", LOG_PERROR | LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	if (daemonize) {
		/* Create a new SID for the child process */
		pid_t sid = setsid();
		if (sid < 0) {
			syslog(LOG_ERR, "Did not get session id for process\n");
			exit(EXIT_FAILURE);
		}

		/* Change the current working directory */
		if (chdir("/var/log/") < 0) {
			syslog(LOG_ERR, "Could not change to directory /var/log/ ...(%s)\n", strerror(errno));
			exit(EXIT_FAILURE);
		}

		/* Close out the standard file descriptors */
		FILE *res = freopen("/dev/null", "r", stdin);
		res = freopen("/dev/null", "w", stdout);
		res = freopen("/dev/null", "w", stderr);
		UNREFERENCED_PARAMETER(res);

		/* Daemon-specific initialization goes here */
		pidfd = open(pidfile, O_RDWR | O_CREAT, 0600);
		if (-1 == pidfd) {
			syslog(LOG_ERR, "Could not open PID lock file %s, exiting\n", pidfile);
			exit(EXIT_FAILURE);
		}
		/* Try to lock file */
		if (lockf(pidfd, F_TLOCK, 0) == -1) {
			syslog(LOG_ERR, "Could not lock PID lock file %s, exiting\n", pidfile);
			exit(EXIT_FAILURE);
		}

		/* Get and format PID */
		char str[10];
		snprintf(str, sizeof(str), "%d\n", getpid());

		/* write pid to lockfile */
		write(pidfd, str, strlen(str));
	}

	int clientFd;
	socklen_t adrlen = sizeof (struct sockaddr_un);
	struct sockaddr_un clientAddress;

	syslog(LOG_INFO, "Successfully initialized\n");

	while (1) {
		clientFd = accept(sockfd, (struct sockaddr *) &clientAddress, &adrlen);
		if (clientFd < 0) {
			if (Debug)
				syslog(LOG_INFO, "Continue %d \n", clientFd);
			continue;
		}
		syslog(LOG_INFO, "New client connected.\n");
		int errorTh = pthread_create(&tid, NULL, &serviceClient, &clientFd);
		if (Debug)
			syslog(LOG_INFO, "pthread_create: %d", errorTh);
	}

	shutdownDaemon(EXIT_SUCCESS);
}
