#include "isegcommon.h"
#include "kdatacache.h"
#include "kloggerfifo.h"
#include "receivethread.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <linux/can/raw.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>


typedef struct KReceiver {
	int socket; /* can raw socket */
	struct msghdr msg;
	struct iovec iov;
	struct can_frame frame;
	struct sockaddr_can addr;
	char ctrlmsg[CMSG_SPACE(sizeof(struct timeval)) + CMSG_SPACE(sizeof(__u32))];
	int count;
} KReceiver;

static KReceiver m_receiver[DataCacheMaxLines] = { 0 };

extern CrateCache m_crateCacheValue[DataCacheMaxLines][DataCacheMaxCrates];

int receivethread_init(const char *canInterface)
{
	int line = atoi(&canInterface[3]);

	if ((m_receiver[line].socket = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		fprintf(stderr, "receive: error socket()\n");
		return 0;
	}

	struct ifreq ifr;
	memset(&ifr.ifr_name, 0, sizeof(ifr.ifr_name));
	strcpy(ifr.ifr_name, canInterface);

	if (ioctl(m_receiver[line].socket, SIOCGIFINDEX, &ifr) < 0) {
		fprintf(stderr, "receive: error SIOCGIFINDEX\n");
		return 0;
	}

	m_receiver[line].addr.can_family = AF_CAN;
	m_receiver[line].addr.can_ifindex = ifr.ifr_ifindex;

	if (bind(m_receiver[line].socket, (struct sockaddr *)&m_receiver[line].addr, sizeof(m_receiver[line].addr)) < 0) {
		fprintf(stderr, "receive: error bind\n");
		return 0;
	}

	m_receiver[line].iov.iov_base = &m_receiver[line].frame;
	m_receiver[line].msg.msg_name = &m_receiver[line].addr;
	m_receiver[line].msg.msg_iov = &m_receiver[line].iov;
	m_receiver[line].msg.msg_iovlen = 1;
	m_receiver[line].msg.msg_control = &m_receiver[line].ctrlmsg;

	return 1;
}

static KCanMessage toKCanMessage(struct can_frame frame, int line)
{
	enum { BitDcp = 0x80 };

	KCanMessage result = EmptyCanMessage;

	const int size = frame.can_dlc;
	result.id.w = frame.can_id;
	result.id.bit.addressOffset |= line;
	result.bit.dlc = frame.can_dlc;

	if (size >= 1 && (frame.data[0] & BitDcp)) {
		result.bit.edcp = false;
		result.dataId = 0x8000 + (frame.data[0] & 0xFF);
		memcpy(result.data, &frame.data[1], size - 1);
	} else if (size >= 2 && !(frame.data[0] & BitDcp)) {
		result.bit.edcp = true;
		result.dataId = (frame.data[0] << 8) + frame.data[1];
		memcpy(result.data, &frame.data[2], size - 2);
	}

	result.bit.busy = true;
	return result;
}

KCanMessage receivethread_waitForMessage(void)
{
	const KCanMessage emptyResult = EmptyCanMessage;
	fd_set rdfs;
	int line = 0;
	int highestDescriptor = 0;

	FD_ZERO(&rdfs);
	for (line = 0; line < DataCacheMaxLines; line++) {
		if (m_receiver[line].socket) {
			FD_SET(m_receiver[line].socket, &rdfs);
			if (highestDescriptor < m_receiver[line].socket)
				highestDescriptor = m_receiver[line].socket;
		}
	}

	struct timeval timeout = { 1, 0 };
	int ret;
	if ((ret = select(highestDescriptor + 1, &rdfs, NULL, NULL, &timeout)) < 0) {
		fprintf(stderr, "receive: error select\n");
		return emptyResult;
	}

	for (line = 0; line < DataCacheMaxLines; line++) {
		if (!FD_ISSET(m_receiver[line].socket, &rdfs)) {
			continue;
		}

		/* these settings may be modified by recvmsg() */
		m_receiver[line].iov.iov_len = sizeof(m_receiver[line].frame);
		m_receiver[line].msg.msg_namelen = sizeof(m_receiver[line].addr);
		m_receiver[line].msg.msg_controllen = sizeof(&m_receiver[line].ctrlmsg);
		m_receiver[line].msg.msg_flags = 0;

		memset(&m_receiver[line].frame, 0, sizeof(struct can_frame));
		int nbytes = recvmsg(m_receiver[line].socket, &m_receiver[line].msg, 0);
		if (nbytes < 0) {
			fprintf(stderr, "receive: error nothing to read\n");
			return emptyResult;
		}

		if (nbytes < (int)sizeof(struct can_frame)) {
			fprintf(stderr, "receive: error incomplete CAN frame\n");
			return emptyResult;
		}

		const KCanMessage canMessage = toKCanMessage(m_receiver[line].frame, line);

		if (m_receiver[line].frame.can_id & 4) {
			bool boardB = 0;
			uint8_t device = ((!canMessage.id.bit.crate) & m_crateCacheValue[0][0].backPlaneType & (1 << line)) ?
					kcan_moduleAddressToSlotNumber(line, canMessage.id.bit.address, &boardB) : canMessage.id.bit.address;
			uint32_t answers = kcache_remainingAnswers(line, device, canMessage.id.bit.crate);
			kloggerfifo_setFrame(LogCanReceive, "CANRX",
								 &canMessage, ++m_receiver[line].count, answers);
		}
		return canMessage;
	}

	return emptyResult;
}
