#include "sendthread.h"

#include <kcan.h>
#include <isegcommon.h>
#include <kdatacache.h>
#include <kdatacache.h>
#include <kloggerfifo.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <unistd.h>

#include <linux/can.h>
#include <linux/can/raw.h>

enum { Debug = 0 };

typedef struct KTransmitter {
	int socket; /* can raw socket */
	struct sockaddr_can addr;
	struct ifreq ifr;
} KTransmitter;

static KTransmitter m_transmitter[DataCacheMaxLines] = { 0 };

void writeLogMessage(const char *messageType, const struct can_frame *frame, int count, uint32_t answers);

extern CrateCache m_crateCacheValue[DataCacheMaxLines][DataCacheMaxCrates];

int sendthread_init(const char *canInterface)
{
	int line = atoi(&canInterface[3]);

	/* open socket */
	if ((m_transmitter[line].socket = socket(PF_CAN, SOCK_RAW, CAN_RAW)) < 0) {
		fprintf(stderr, "send: error opening socket\n");
		return 0;
	}

	m_transmitter[line].addr.can_family = AF_CAN;

	strcpy(m_transmitter[line].ifr.ifr_name, canInterface);
	if (ioctl(m_transmitter[line].socket, SIOCGIFINDEX, &m_transmitter[line].ifr) < 0) {
		fprintf(stderr, "send: error SIOCGIFINDEX\n");
		return 0;
	}
	m_transmitter[line].addr.can_ifindex = m_transmitter[line].ifr.ifr_ifindex;

	/* disable default receive filter on this RAW socket */
	/* This is obsolete as we do not read from the socket at all, but for */
	/* this reason we can remove the receive list in the Kernel to save a */
	/* little (really a very little!) CPU usage.                          */
	setsockopt(m_transmitter[line].socket, SOL_CAN_RAW, CAN_RAW_FILTER, NULL, 0);

	// disable loopback, otherwise all frames immediately occur in the receive thread
	int loopback = 0;
	setsockopt(m_transmitter[line].socket, SOL_CAN_RAW, CAN_RAW_LOOPBACK, &loopback, sizeof(loopback));

	if (bind(m_transmitter[line].socket, (struct sockaddr *)&m_transmitter[line].addr,
			 sizeof(m_transmitter[line].addr)) < 0) {
		fprintf(stderr, "send: error bind\n");
		return 0;
	}

	return 1;
}

static struct can_frame toCanFrame(KCanMessage *msg)
{
	const int size = msg->bit.dlc;

	struct can_frame frame = { 0 };
	frame.can_id = msg->id.w;
	frame.can_id |= msg->bit.rtr << 30;
	frame.can_id |= msg->bit.extendedId << 31;
	frame.can_dlc = size;
	if (msg->bit.edcp && size >= 2) {
		frame.data[0] = msg->dataId >> 8;
		frame.data[1] = msg->dataId & 0xFF;
		memcpy(&frame.data[2], msg->data, size - 2);
	} else if (!msg->bit.edcp && size >= 1) {
		frame.data[0] = msg->dataId & 0xFF;
		memcpy(&frame.data[1], msg->data, size - 1);
	}

	return frame;
}

/******************************************************************************
 * Master/Slave Crate CC44A muss immer mit dem Communicator Linie can0
 * verbunden sein!
 *
 * Multi-Interfacesystem:
 * Arbeitsverbot auf nicht konnektierten Linien
 ******************************************************************************/
void sendthread_send(KCanMessage *msg)
{
	static bool communicationLost = false;
	static int counterTx = 0;
	static int counterRq = 0;
	int line = msg->id.bit.addressOffset;

	if (!m_transmitter[0].socket && !m_transmitter[line].socket) {
		//kein Socket
		return;
	}

	if (line && m_transmitter[line].socket) {
		msg->id.bit.addressOffset = 0;
		msg->bit.extendedId = 0;
	} else {
		line = 0;
	}

	struct can_frame frame = toCanFrame(msg);
	uint32_t logLevel = kdatacache_getSystemValue(DataCacheColumnSystemLogLevel);
	bool boardB = 0;
	uint8_t device = ((!msg->id.bit.crate) & m_crateCacheValue[0][0].backPlaneType & (1 << line)) ?
			kcan_moduleAddressToSlotNumber(line, msg->id.bit.address, &boardB) : msg->id.bit.address;
	uint32_t answers = kcache_remainingAnswers(line, device, msg->id.bit.crate);
	bool readRequest = msg->id.bit.readRequest;

	char *loggerInfo;
	while (kloggerfifo_getInfo(&loggerInfo)) {
		writeLogMessage(loggerInfo, NULL, ++counterTx, answers);
	}

	if (logLevel & LogCanWriteRequest && (!readRequest) ) {
		writeLogMessage("80 CANTX", &frame, ++counterTx, answers);
	} else if (logLevel & LogCanReadRequest && readRequest) {
		writeLogMessage("40 CANRQ", &frame, ++counterRq, answers);
	}

	int nbytes;
	if ((nbytes = write(m_transmitter[line].socket, &frame, sizeof(frame))) != sizeof(frame)) {
		if (!communicationLost) {
			fprintf(stderr, "send: write error!");
			kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanStatus);
			communicationLost = true;
		}
		return;
	}

	if (communicationLost)
		kdatacache_createSystemTimeStamp(true, DataCacheColumnSystemCanStatus);

	communicationLost = false;
}
