#include <unity_fixture.h>

#include "../isegapi.h"
#include "../kdatacache.h"
#include "../kcanapi.h"
#include "../version.h"

#include <kcan.h>
#include <kedcp_can_dataid.h>

#include <stdlib.h>
#include <string.h>

TEST_GROUP(kdatacache);

void kmodule_setAlive(bool alive, uint16_t line, uint16_t module);
void kcrate_setAlive(bool alive, uint16_t line, uint16_t crate);
void kcrateEch238_setAlive(bool alive, uint16_t line, uint16_t crate);
float kdatacache_getChannelValueSetR4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column);
uint32_t kdatacache_getChannelValueSetUI4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column, uint8_t index);
float kdatacache_getModuleValueSetR4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index);
uint32_t kdatacache_getModuleValueSetUI4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index);
const char *kdatacache_getModuleValueSetString(uint16_t line, uint16_t module, KDataCacheModuleColumn column);

extern uint32_t m_timeOutBackPlaneType;

TEST_SETUP(kdatacache)
{
	kdatacache_clear();
	m_timeOutBackPlaneType = TIME_OUT_MODULE_LOG_ON;
}

TEST_TEAR_DOWN(kdatacache)
{
}

TEST(kdatacache, clear)
{
	kdatacache_setChannelValueUI4(0x12345678, 0, 0, 0, DataCacheColumnChannelStatus, NO_BIT_ARRAY_INDEX);

	kdatacache_clear();

	TEST_ASSERT_EQUAL(0, kdatacache_getChannelValueUI4(0, 0, 0, DataCacheColumnChannelStatus, NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, getValue)
{
	TEST_ASSERT_EQUAL(0, kdatacache_getChannelValueUI4(0, 0, 0, 0, NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, getVersion)
{
	const unsigned int version[] = { VER_FILEVERSION };
	const unsigned int expected =
		(version[0] << 24) | (version[1] << 16 | (version[2] << 8) | (version[3]));
	TEST_ASSERT_EQUAL_HEX32(expected, can_getVersion());
}

TEST(kdatacache, getVersionString)
{
	TEST_ASSERT_EQUAL_STRING(VER_FILEVERSION_STR, can_getVersionString());
}

static KCanMessage buildCanMessageUI4(KEdcpDataId dataId, uint8_t channel, uint8_t dlc, uint32_t data)
{
	KCanMessage msg = EmptyCanMessage;
	uint8_t index = 0;
	msg.dataId = dataId;
	msg.bit.dlc = dlc;
	if (dataId & EDCP_DATAID_CHANNEL_CODE) {
		msg.data[index++] = channel;
	}

	msg.data[index++] = (data >> 24) & 0xFF;
	msg.data[index++] = (data >> 16) & 0xFF;
	msg.data[index++] = (data >>  8) & 0xFF;
	msg.data[index] = (data >>  0) & 0xFF;

	return msg;
}

static KCanMessage buildCanMessageUI2(KEdcpDataId dataId, uint8_t channel, uint8_t dlc, uint32_t data)
{
	KCanMessage msg = EmptyCanMessage;
	uint8_t index = 0;
	msg.dataId = dataId;
	msg.bit.dlc = dlc;
	if (dataId & EDCP_DATAID_CHANNEL_CODE) {
		msg.data[index++] = channel;
	}

	msg.data[index++] = (data >>  8) & 0xFF;
	msg.data[index] = (data >>  0) & 0xFF;

	return msg;
}

static KCanMessage buildCanMessageString(KEdcpDataId dataId, uint8_t dlc, const char *data)
{
	KCanMessage msg = EmptyCanMessage;

	msg.dataId = dataId;
	msg.bit.dlc = dlc;
	strcpy((char *)msg.data, data);

	return msg;
}

static KCanMessage buildCanMessageR4(KEdcpDataId dataId, uint8_t channel, uint8_t dlc, float data)
{
	uint32_t value;
	memcpy(&value, &data, 4);
	return buildCanMessageUI4(dataId, channel, dlc, value);
}

TEST(kdatacache, setCanMessageValue16)
{
	enum { iseg = 0x15E6 };
	KCanMessage msg = buildCanMessageUI4(EDCP_DATAID_CHN_STATUS, 1, 5, iseg << 16);

	kdatacache_setCanMessageValue(&msg);

	TEST_ASSERT_EQUAL_HEX16(iseg, kdatacache_getChannelValueUI4(0, 0, 1,
																DataCacheColumnChannelStatus, NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, setCanMessageValue32)
{
	enum { test = 0x12345678 };
	KCanMessage msg = buildCanMessageUI4(EDCP_DATAID_CHN_VOLTAGE_SET, DataCacheMaxChannels - 1, 7, test);

	kdatacache_setCanMessageValue(&msg);

	TEST_ASSERT_EQUAL_HEX32(test, kdatacache_getChannelValueUI4(0, 0, DataCacheMaxChannels - 1,
																DataCacheColumnChannelVoltageSet, NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, setCanMessageValueR4)
{
	float vset = 1000.1234;
	KCanMessage msg = buildCanMessageR4(EDCP_DATAID_CHN_VOLTAGE_SET, DataCacheMaxChannels - 1, 7, vset);

	kdatacache_setCanMessageValue(&msg);

	TEST_ASSERT_EQUAL_FLOAT(vset, kdatacache_getChannelValueR4(0, 0, DataCacheMaxChannels - 1, DataCacheColumnChannelVoltageSet));
}

TEST(kdatacache, setChannelValueR4)
{
	kdatacache_setChannelValueR4(1234.5678, 0, 0, 2, DataCacheColumnChannelVoltageSet);

	kdatacache_clear();

	TEST_ASSERT_EQUAL(0, kdatacache_getChannelValueR4(0, 0, 2, DataCacheColumnChannelVoltageSet));
}

TEST(kdatacache, setItemValueVset)
{
	IsegItem item = { "0.1.47.VoltageSet", "2345.67", "000", "0", "" };
	float vset = 2345.67;

	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_OK, result);

	TEST_ASSERT_EQUAL_FLOAT(vset, kdatacache_getChannelValueSetR4(0, 1, DataCacheMaxChannels - 1, DataCacheColumnChannelVoltageSet));
}

TEST(kdatacache, setItemValueVsetFail1)
{
	IsegItem item = { "0.1.VoltageSet", "2000", "000", "0", "" };
	float vset = 0;

	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_ERROR, result);

	TEST_ASSERT_EQUAL_FLOAT(vset, kdatacache_getChannelValueSetR4(0, 1, DataCacheMaxChannels - 1, DataCacheColumnChannelVoltageSet));
}

TEST(kdatacache, setItemValueVsetFail2)
{
	IsegItem item = { "0.VoltageSet", "2000", "000", "0", "" };
	float vset = 0;

	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_ERROR, result);

	TEST_ASSERT_EQUAL_FLOAT(vset, kdatacache_getChannelValueSetR4(0, 1, DataCacheMaxChannels - 1, DataCacheColumnChannelVoltageSet));
}

TEST(kdatacache, setItemValueVsetFail3)
{
	IsegItem item = { "0.1.48.VoltageSet", "2000", "000", "0", "" };
	float vset = 0;

	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_ERROR, result);

	TEST_ASSERT_EQUAL_FLOAT(vset, kdatacache_getChannelValueSetR4(0, 1, DataCacheMaxChannels - 1, DataCacheColumnChannelVoltageSet));
}

TEST(kdatacache, setItemValueIset)
{
	IsegItem item = { "0.1.47.CurrentSet", "0.00123", "000", "0", "" };

	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_OK, result);

	float iset = kdatacache_getChannelValueSetR4(0, 1, DataCacheMaxChannels - 1, DataCacheColumnChannelCurrentSet);
	TEST_ASSERT_EQUAL_FLOAT(0.00123, iset);
}

TEST(kdatacache, setItemValueOn)
{
	IsegItem item = { "0.2.1.Control", "8", "000", "0", "" };

	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_OK, result);

	uint32_t on = kdatacache_getChannelValueSetUI4(0, 2, 1, DataCacheColumnChannelControl, NO_BIT_ARRAY_INDEX);
	TEST_ASSERT_EQUAL_INT32(8, on);
}

TEST(kdatacache, setItemValueVoltageRampSpeed)
{
	IsegItem item = { "0.1.VoltageRampSpeed", "11.111", "000", "0", "" };

	kmodule_setAlive(true, 0, 1);
	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_OK, result);

	float vramp = kdatacache_getModuleValueSetR4(0, 1, DataCacheColumnModuleVoltageRampSpeed, 0);
	TEST_ASSERT_EQUAL_FLOAT(11.111, vramp);
}

TEST(kdatacache, setItemValueCurrentRampSpeed)
{
	IsegItem item = { "0.1.CurrentRampSpeed", "51.111", "000", "0", "" };

	kmodule_setAlive(true, 0, 1);
	IsegResult result = can_setItemValue(&item);

	TEST_ASSERT_EQUAL_INT32(ISEG_OK, result);

	float iramp = kdatacache_getModuleValueSetR4(0, 1, DataCacheColumnModuleCurrentRampSpeed, 0);
	TEST_ASSERT_EQUAL_FLOAT(51.111, iramp);
}

TEST(kdatacache, getItemValueVoltageMeasure)
{
	const float vmeas = 1000.1234;
	const char object[128] = { "0.63.47.VoltageMeasure" };
	KCanMessage msg = buildCanMessageR4(EDCP_DATAID_CHN_VOLTAGE_MEASURE, DataCacheMaxChannels - 1, 7, vmeas);
	msg.id.bit.address = 63;

	kdatacache_setCanMessageValue(&msg);
	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(vmeas, atof(item.value));
}

TEST(kdatacache, getItemValueVoltageNominal)
{
	const float vnom = -10000;
	const char object[128] = { "0.0.0.VoltageNominal" };
	KCanMessage msg = buildCanMessageR4(EDCP_DATAID_CHN_VOLTAGE_NOMINAL, 0, 7, vnom);
	msg.id.bit.address = 0;

	kdatacache_setCanMessageValue(&msg);
	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(vnom, atof(item.value));
}

TEST(kdatacache, getItemValueChannelEventMask)
{
	const char eventMask[] = "0.0.2.EventMask";
	const int mask = 0x1234;
	const KCanMessage msg = buildCanMessageUI2(EDCP_DATAID_CHN_EVENT_MASK, 2, 5, mask);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(eventMask);

	TEST_ASSERT_EQUAL_STRING(eventMask, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_HEX16(mask, atoi(item.value));
}

TEST(kdatacache, setItemValueChannelEventMask)
{
	const IsegItem item = { "0.4.2.EventMask", "22136", "", "", "" };

	TEST_ASSERT_EQUAL_INT32(ISEG_OK, can_setItemValue(&item));
	TEST_ASSERT_EQUAL(22136, kdatacache_getChannelValueSetUI4(0, 4, 2, DataCacheColumnChannelEventMask,
														   NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, getItemValueVoltageRampSpeed)
{
	float vramp = 12.345;
	KCanMessage msg = buildCanMessageR4(EDCP_DATAID_MOD_VOLTAGE_RAMP_SPEED, 0, 6, vramp);
	msg.id.bit.address = 1;

	kdatacache_setCanMessageValue(&msg);
	const IsegItem item = can_getItemValue("0.1.VoltageRampSpeed");

	TEST_ASSERT_EQUAL_STRING("0.1.VoltageRampSpeed", item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(vramp, atof(item.value));
}

TEST(kdatacache, getItemValueCurrentRampSpeed)
{
	float iramp = 51.345;
	KCanMessage msg = buildCanMessageR4(EDCP_DATAID_MOD_CURRENT_RAMP_SPEED, 0, 6, iramp);
	msg.id.bit.address = 1;

	kdatacache_setCanMessageValue(&msg);
	const IsegItem item = can_getItemValue("0.1.CurrentRampSpeed");

	TEST_ASSERT_EQUAL_STRING("0.1.CurrentRampSpeed", item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(iramp, atof(item.value));
}

TEST(kdatacache, getItemValueVoltageLimit)
{
	const char voltageLimit[] = "0.0.VoltageLimit";
	const float limit = 80.f;
	const KCanMessage msg = buildCanMessageR4(EDCP_DATAID_MOD_VOLTAGE_LIMIT_REL, 0, 6, limit);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(voltageLimit);

	TEST_ASSERT_EQUAL_STRING(voltageLimit, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(limit, atof(item.value));
}

TEST(kdatacache, getItemValueCurrentLimit)
{
	const char currentLimit[] = "0.0.CurrentLimit";
	const float limit = 102.f;
	const KCanMessage msg = buildCanMessageR4(EDCP_DATAID_MOD_CURRENT_LIMIT_REL, 0, 6, limit);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(currentLimit);

	TEST_ASSERT_EQUAL_STRING(currentLimit, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(limit, atof(item.value));
}

TEST(kdatacache, getItemValueSupplyP5)
{
	const char supplyP5[] = "0.0.Supplies[2]";
	const float supply = 5.01f;
	const KCanMessage msg = buildCanMessageR4(EDCP_DATAID_MOD_SUPPLY5, 0, 6, supply);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(supplyP5);

	TEST_ASSERT_EQUAL_STRING(supplyP5, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(supply, atof(item.value));
}

TEST(kdatacache, getItemValueSupplyP24)
{
	const char supplyP24[] = "0.0.Supplies[0]";
	const float supply = 23.9f;
	const KCanMessage msg = buildCanMessageR4(EDCP_DATAID_MOD_SUPPLY24, 0, 6, supply);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(supplyP24);

	TEST_ASSERT_EQUAL_STRING(supplyP24, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(supply, atof(item.value));
}

TEST(kdatacache, getItemValue_moduleSupplyOutOfRange)
{
	const char supplyOutOfRange[] = "0.0.Supplies[9]";

	const IsegItem item = can_getItemValue(supplyOutOfRange);

	TEST_ASSERT_EQUAL_STRING(supplyOutOfRange, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_ERROR, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(0.f, atof(item.value));
}

TEST(kdatacache, getItemValue_crateSupplyOutOfRange)
{
	const char tempOutOfRange[] = "0.1000.Supplies[9]";

	const IsegItem item = can_getItemValue(tempOutOfRange);

	TEST_ASSERT_EQUAL_STRING(tempOutOfRange, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_ERROR, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(0.f, atof(item.value));
}

TEST(kdatacache, getItemValue_moduleTemperatureOutOfRange)
{
	const char tempOutOfRange[] = "0.0.Temperatures[9]";

	const IsegItem item = can_getItemValue(tempOutOfRange);

	TEST_ASSERT_EQUAL_STRING(tempOutOfRange, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_ERROR, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(0.f, atof(item.value));
}

TEST(kdatacache, getItemValue_crateTemperatureOutOfRange)
{
	const char supplyOutOfRange[] = "0.1000.Temperatures[9]";

	const IsegItem item = can_getItemValue(supplyOutOfRange);

	TEST_ASSERT_EQUAL_STRING(supplyOutOfRange, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_ERROR, item.quality);
	TEST_ASSERT_EQUAL_FLOAT(0.f, atof(item.value));
}

TEST(kdatacache, getItemValueEventMask)
{
	const char eventMask[] = "0.0.EventMask";
	const int mask = 0x1234;
	const KCanMessage msg = buildCanMessageUI2(EDCP_DATAID_MOD_EVENT_MASK, 0, 4, mask);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(eventMask);

	TEST_ASSERT_EQUAL_STRING(eventMask, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_HEX16(mask, atoi(item.value));
}

TEST(kdatacache, setItemValueEventMask)
{
	const IsegItem item = { "0.4.EventMask", "22136", "", "", "" };

	kmodule_setAlive(true, 0, 4);
	TEST_ASSERT_EQUAL_INT32(ISEG_OK, can_setItemValue(&item));
	TEST_ASSERT_EQUAL(22136, kdatacache_getModuleValueSetUI4(0, 4, DataCacheColumnModuleEventMask,
															 NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, getItemValueSampleRate)
{
	const char sampleRate[] = "0.0.SampleRate";
	const int samples = 500;
	const KCanMessage msg = buildCanMessageUI2(EDCP_DATAID_MOD_ADC_SAMPLES_PER_SECOND, 0, 4, samples);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(sampleRate);

	TEST_ASSERT_EQUAL_STRING(sampleRate, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL(samples, atoi(item.value));
}

TEST(kdatacache, setItemValueSampleRate)
{
	const IsegItem item = { "0.4.SampleRate", "50", "", "", "" };

	kmodule_setAlive(true, 0, 4);
	TEST_ASSERT_EQUAL_INT32(ISEG_OK, can_setItemValue(&item));
	TEST_ASSERT_EQUAL(50, kdatacache_getModuleValueSetUI4(0, 4, DataCacheColumnModuleSampleRate,
													   NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, getItemValueDigitalFilter)
{
	const char digitalFilter[] = "0.0.DigitalFilter";
	const int filter = 256;
	const KCanMessage msg = buildCanMessageUI2(EDCP_DATAID_MOD_DIGITAL_FILTER, 0, 4, filter);
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(digitalFilter);

	TEST_ASSERT_EQUAL_STRING(digitalFilter, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL(filter, atoi(item.value));
}

TEST(kdatacache, setItemValueDigitalFilter)
{
	const IsegItem item = { "0.4.DigitalFilter", "1024", "", "", "" };

	kmodule_setAlive(true, 0, 4);
	TEST_ASSERT_EQUAL(ISEG_OK, can_setItemValue(&item));
	TEST_ASSERT_EQUAL(1024, kdatacache_getModuleValueSetUI4(0, 4, DataCacheColumnModuleDigitalFilter,
															NO_BIT_ARRAY_INDEX));
}

TEST(kdatacache, getItemValueOn)
{
	uint16_t on = 8;
	const char object[128] = { "0.1.15.Control" };
	KCanMessage msg = buildCanMessageUI2(EDCP_DATAID_CHN_CONTROL, 15, 5, on);
	msg.id.bit.address = 1;

	kdatacache_setCanMessageValue(&msg);
	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_UINT16(on, atoi(item.value));
}

TEST(kdatacache, getItemValueAlive)
{
	enum { Module = 0 };
	const char object[] = "0.0.Alive";
	kmodule_setAlive(true, 0, Module);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("1", item.value);
}

TEST(kdatacache, getItemValueModuleNumber_afterInit)
{
	const char object[] = "0.ModuleNumber";

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_INITIALIZE, item.quality);
	TEST_ASSERT_EQUAL_STRING("0", item.value);
}

TEST(kdatacache, getItemValueModuleNumber_oneModuleConnected)
{
	const char object[] = "0.ModuleNumber";
	kmodule_setAlive(true, 0, 0);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("1", item.value);
}

TEST(kdatacache, getItemValueModuleNumber_noModuleConnected)
{
	const char object[] = "0.ModuleNumber";
	kmodule_setAlive(true, 0, 0);
	kmodule_setAlive(false, 0, 0);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("0", item.value);
}

TEST(kdatacache, getItemValueModuleNumber_minimumIsZero)
{
	const char object[] = "0.ModuleNumber";
	kmodule_setAlive(true, 0, 0);
	kmodule_setAlive(false, 0, 0);
	kmodule_setAlive(false, 0, 0);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("0", item.value);
}

TEST(kdatacache, getItemModuleList_empty)
{
	const char object[] = "0.ModuleList";

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_INITIALIZE, item.quality);
	TEST_ASSERT_EQUAL_STRING("", item.value);
}

TEST(kdatacache, getItemModuleListOk)
{
	const char object[] = "0.ModuleList";
	kmodule_setAlive(true, 0, 0);
	kmodule_setAlive(true, 0, 5);
	kmodule_setAlive(true, 0, 17);
	kmodule_setAlive(true, 0, 63);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("0,5,17,63", item.value);
}

TEST(kdatacache, getItemModuleList_maximumNumber)
{
	const char object[] = "0.ModuleList";
	int i;

	for (i = 0; i < DataCacheMaxModules; ++i)
		kmodule_setAlive(true, 0, i);

	const char expectedList[] =
		"0,1,2,3,4,5,6,7,8,9,"
		"10,11,12,13,14,15,16,17,18,19,"
		"20,21,22,23,24,25,26,27,28,29,"
		"30,31,32,33,34,35,36,37,38,39,"
		"40,41,42,43,44,45,46,47,48,49,"
		"50,51,52,53,54,55,56,57,58,59,"
		"60,61,62,63";

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING(expectedList, item.value);
}

TEST(kdatacache, getItemModuleListNotOk)
{
	const char object[] = "0.ModuleList";
	kmodule_setAlive(true, 0, 0);
	kmodule_setAlive(true, 0, 5);
	kmodule_setAlive(true, 0, 64);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("0,5", item.value);
}

TEST(kdatacache, getItemCrateList_ok)
{
	const char object[] = "0.CrateList";
	kcrate_setAlive(true, 0, 0);
	kcrate_setAlive(true, 0, 15);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("1000,1015", item.value);
}

TEST(kdatacache, getItemCrateList_maximum)
{
	const char object[] = "0.CrateList";
	int i;

	for (i = 0; i < DataCacheMaxCrates; ++i)
		kcrate_setAlive(true, 0, i);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("1000,1001,1002,1003,1004,1005,"
							 "1006,1007,1008,1009,1010,1011,"
							 "1012,1013,1014,1015", item.value);
}

TEST(kdatacache, getItemCrateList_notOk)
{
	const char object[] = "0.CrateList";
	kcrate_setAlive(true, 0, DataCacheMaxCrates);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("", item.value);
}

TEST(kdatacache, getItemModuleHighVoltageOkInit)
{
	const char object[] = "0.4.HighVoltageOk";

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_INITIALIZE, item.quality);
	TEST_ASSERT_EQUAL_STRING("", item.value);
}

TEST(kdatacache, getItemModuleHighVoltageOk)
{
	const char object[] = "0.0.HighVoltageOk";
	KCanMessage msg = buildCanMessageString(EDCP_DATAID_MOD_CONFIGURE_HIGH_VOLTAGE_OK, 8, "noHVOK");
	kdatacache_setCanMessageValue(&msg);

	const IsegItem item = can_getItemValue(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("noHVOK", item.value);
}

TEST(kdatacache, setItemModuleHighVoltageOk)
{
	const IsegItem item = { "0.4.HighVoltageOk", "cgHVOK", "", "", "" };

	kmodule_setAlive(true, 0, 4);
	TEST_ASSERT_EQUAL(ISEG_OK, can_setItemValue(&item));
	TEST_ASSERT_EQUAL_STRING("cgHVOK",
							 kdatacache_getModuleValueSetString(0, 4,
																DataCacheColumnModuleHighVoltageOk));
}

TEST(kdatacache, getItemPropertyVset)
{
	const char object[128] = { "0.1.15.VoltageSet" };
	IsegItemProperty item = can_getItemProperty(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING("R4", item.type);
	TEST_ASSERT_EQUAL_STRING("V", item.unit);
	TEST_ASSERT_EQUAL_STRING("RW", item.access);
}

TEST(kdatacache, getItemPropertyFirmwareName)
{
	const char object[128] = { "0.1.FirmwareName" };
	IsegItemProperty item = can_getItemProperty(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING("STR", item.type);
	TEST_ASSERT_EQUAL_STRING("", item.unit);
	TEST_ASSERT_EQUAL_STRING("R", item.access);
}

TEST(kdatacache, getItemPropertyAlive)
{
	const char object[] = "0.0.Alive";
	IsegItemProperty item = can_getItemProperty(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING("BOOL", item.type);
	TEST_ASSERT_EQUAL_STRING("", item.unit);
	TEST_ASSERT_EQUAL_STRING("R", item.access);
}

TEST(kdatacache, getItemPropertyModuleNumber)
{
	const char object[] = "0.ModuleNumber";
	const IsegItemProperty item = can_getItemProperty(object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING("UI4",  item.type);
	TEST_ASSERT_EQUAL_STRING("",     item.unit);
	TEST_ASSERT_EQUAL_STRING("R",    item.access);
}

KCanMessage canMessageSystemValueString(char *string);
TEST(kdatacache, tunnelMessageSystemValueString)
{
	const uint8_t hexArray[8] = { 0x12, 0xaa, 0x4f, 0x0, 0x0, 0x0, 0x0, 0x0 };
	const KCanMessage msg = canMessageSystemValueString("004#12aa4f");
	TEST_ASSERT_FALSE(msg.bit.edcp);
	TEST_ASSERT_EQUAL_INT(0x004, msg.id.w);
	TEST_ASSERT_EQUAL_HEX8(3, msg.bit.dlc);
	TEST_ASSERT_EQUAL_HEX8(hexArray[0], msg.dataId);
	TEST_ASSERT_EQUAL_HEX8_ARRAY(&hexArray[1], msg.data, 2);
}

TEST(kdatacache, getItemCratePowerOn)
{
	const char name[] = "test";
	const char object[] = "8.2000.PowerOn";
	const int power = 0x0000;
	uint16_t line = 8;
	KCanMessage msg = buildCanMessageUI2(DCP_DATAID + DATAID_GET_POWER_ON_OFF, 0, 3, power);
	msg.id.bit.crate = 1;
	msg.id.bit.addressOffset = line;
	kdatacache_setCanMessageValue(&msg);

	iseg_connect(name, name, NULL);
	kcrateEch238_setAlive(true, line, 0);
	const IsegItem item = iseg_getItem(name, object);

	TEST_ASSERT_EQUAL_STRING(object, item.object);
	TEST_ASSERT_EQUAL_STRING(ISEG_ITEM_QUALITY_OK, item.quality);
	TEST_ASSERT_EQUAL_STRING("1", item.value);
}

TEST(kdatacache, setItemSetCrateEch238PowerOn)
{
	const char name[] = "test";
	const char object[OBJECT_NAME_SIZE] = { "8.2000.PowerOn" };
	const char value[VALUE_SIZE] = { "1" };
	uint16_t line = 8;

	iseg_connect(name, name, NULL);
	kcrateEch238_setAlive(true, line, 0);
	TEST_ASSERT_EQUAL(ISEG_OK, iseg_setItem(name, object, value));
	TEST_ASSERT_EQUAL(1, kdatacache_crateCacheValue(line, 2000 - CRATE_ECH238_ADDRESS_OFFSET, true)->powerOnSet);
}

TEST(kdatacache, setItemSetCrateEch238PowerOnFail)
{
	const char name[] = "test";
	const char object[OBJECT_NAME_SIZE] = { "2000.PowerOn" };
	const char value[VALUE_SIZE] = { "1" };
	uint16_t line = 8;

	iseg_connect(name, name, NULL);
	kcrateEch238_setAlive(true, line, 0);
	TEST_ASSERT_EQUAL(ISEG_ERROR, iseg_setItem(name, object, value));
	TEST_ASSERT_EQUAL(0, kdatacache_crateCacheValue(line, 2000 - CRATE_ECH238_ADDRESS_OFFSET, true)->powerOnSet);
}

TEST(kdatacache, wrongSetItemSetCrateEch238PowerOn)
{
	const char name[] = "test";
	const char object[OBJECT_NAME_SIZE] = { "8.1000.PowerOn" };
	const char value[VALUE_SIZE] = { "1" };
	uint16_t line = 8;

	iseg_connect(name, name, NULL);
	kcrateEch238_setAlive(true, line, 0);
	TEST_ASSERT_EQUAL(ISEG_ERROR, iseg_setItem(name, object, value));
	TEST_ASSERT_FALSE((bool)kdatacache_crateCacheValue(line, 1000 - CRATE_ADDRESS_OFFSET, true)->powerOnSet);
}

TEST_GROUP_RUNNER(kdatacache)
{
	RUN_TEST_CASE(kdatacache, clear);

	RUN_TEST_CASE(kdatacache, getValue);
	RUN_TEST_CASE(kdatacache, getVersion);
	RUN_TEST_CASE(kdatacache, getVersionString);
	RUN_TEST_CASE(kdatacache, setCanMessageValue16);
	RUN_TEST_CASE(kdatacache, setCanMessageValue32);
	RUN_TEST_CASE(kdatacache, setCanMessageValueR4);
	RUN_TEST_CASE(kdatacache, setChannelValueR4);
	RUN_TEST_CASE(kdatacache, setItemValueVset);
	RUN_TEST_CASE(kdatacache, setItemValueVsetFail1);
	RUN_TEST_CASE(kdatacache, setItemValueVsetFail2);
	RUN_TEST_CASE(kdatacache, setItemValueVsetFail3);
	RUN_TEST_CASE(kdatacache, setItemValueIset);
	RUN_TEST_CASE(kdatacache, setItemValueOn);
	RUN_TEST_CASE(kdatacache, getItemValueVoltageMeasure);
	RUN_TEST_CASE(kdatacache, getItemValueVoltageNominal)
	RUN_TEST_CASE(kdatacache, getItemValueChannelEventMask);
	RUN_TEST_CASE(kdatacache, setItemValueChannelEventMask);
	RUN_TEST_CASE(kdatacache, setItemValueVoltageRampSpeed);
	RUN_TEST_CASE(kdatacache, setItemValueCurrentRampSpeed);
	RUN_TEST_CASE(kdatacache, getItemValueVoltageRampSpeed);
	RUN_TEST_CASE(kdatacache, getItemValueCurrentRampSpeed);
	RUN_TEST_CASE(kdatacache, getItemValueVoltageLimit);
	RUN_TEST_CASE(kdatacache, getItemValueCurrentLimit);
	RUN_TEST_CASE(kdatacache, getItemValueSupplyP5);
	RUN_TEST_CASE(kdatacache, getItemValueSupplyP24);
	RUN_TEST_CASE(kdatacache, getItemValue_moduleSupplyOutOfRange);
	RUN_TEST_CASE(kdatacache, getItemValue_crateSupplyOutOfRange);
	RUN_TEST_CASE(kdatacache, getItemValue_moduleTemperatureOutOfRange);
	RUN_TEST_CASE(kdatacache, getItemValue_crateTemperatureOutOfRange);
	RUN_TEST_CASE(kdatacache, getItemValueEventMask);
	RUN_TEST_CASE(kdatacache, setItemValueEventMask);
	RUN_TEST_CASE(kdatacache, getItemValueSampleRate);
	RUN_TEST_CASE(kdatacache, setItemValueSampleRate);
	RUN_TEST_CASE(kdatacache, getItemValueDigitalFilter);
	RUN_TEST_CASE(kdatacache, setItemValueDigitalFilter);
	RUN_TEST_CASE(kdatacache, getItemValueOn)
	RUN_TEST_CASE(kdatacache, getItemValueAlive);
	RUN_TEST_CASE(kdatacache, getItemValueModuleNumber_afterInit);
	RUN_TEST_CASE(kdatacache, getItemValueModuleNumber_oneModuleConnected);
	RUN_TEST_CASE(kdatacache, getItemValueModuleNumber_noModuleConnected);
	RUN_TEST_CASE(kdatacache, getItemValueModuleNumber_minimumIsZero);
	RUN_TEST_CASE(kdatacache, getItemModuleList_empty);
	RUN_TEST_CASE(kdatacache, getItemModuleListOk);
	RUN_TEST_CASE(kdatacache, getItemModuleList_maximumNumber);
	RUN_TEST_CASE(kdatacache, getItemModuleListNotOk);
	RUN_TEST_CASE(kdatacache, getItemCrateList_ok);
	RUN_TEST_CASE(kdatacache, getItemCrateList_maximum);
	RUN_TEST_CASE(kdatacache, getItemCrateList_notOk);
	RUN_TEST_CASE(kdatacache, getItemModuleHighVoltageOkInit);
	RUN_TEST_CASE(kdatacache, getItemModuleHighVoltageOk);
	RUN_TEST_CASE(kdatacache, setItemModuleHighVoltageOk);


	RUN_TEST_CASE(kdatacache, getItemPropertyVset);
	RUN_TEST_CASE(kdatacache, getItemPropertyFirmwareName);
	RUN_TEST_CASE(kdatacache, getItemPropertyAlive);
	RUN_TEST_CASE(kdatacache, getItemPropertyModuleNumber);

	RUN_TEST_CASE(kdatacache, tunnelMessageSystemValueString);

	RUN_TEST_CASE(kdatacache, getItemCratePowerOn);
	RUN_TEST_CASE(kdatacache, setItemSetCrateEch238PowerOn);
	RUN_TEST_CASE(kdatacache, setItemSetCrateEch238PowerOnFail);
	RUN_TEST_CASE(kdatacache, wrongSetItemSetCrateEch238PowerOn);

	// Channel Message
	// Module Message
}
